"""A probabilistic cell hashing demultiplexing method.

HashSolo generates a noise distribution and signal distribution
for each hashing barcode from empirically observed counts.
These distributions are updates from the global signal and noise barcode distributions,
which helps in the setting where not many cells are observed.
For a hashing barcode:

Signal distributions
    are estimated from samples where that hashing barcode has the highest count.

Noise distributions
    are estimated from samples where that hashing barcode is one the k-2 lowest barcodes,
    where k is the number of barcodes.

We test each of the following hypotheses in a bayesian fashion,
and select the most probable hypothesis.

A doublet
    should have its two highest barcode counts most likely
    coming from a signal distribution for those barcodes.

A singlet
    should have its highest barcode from a signal distribution,
    and its second highest barcode from a noise distribution.

A negative two highest barcodes
    should come from noise distributions.
"""

from __future__ import annotations

from itertools import product
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from scipy.stats import norm

from ..._compat import old_positionals
from ..._utils import check_nonnegative_integers
from ..._utils._doctests import doctest_skip

if TYPE_CHECKING:
    from collections.abc import Sequence

    from anndata import AnnData
    from numpy.typing import ArrayLike, NDArray


def _calculate_log_likelihoods(  # noqa: PLR0915
    data: np.ndarray, number_of_noise_barcodes: int
) -> tuple[NDArray[np.float64], NDArray[np.float64], dict[int, str]]:
    """Calculate log likelihoods for each hypothesis, negative, singlet, doublet.

    Parameters
    ----------
    data
        cells by hashing counts matrix
    number_of_noise_barcodes
        number of barcodes to used to calculated noise distribution

    Returns
    -------
    log_likelihoods_for_each_hypothesis
        a 2d np.array log likelihood of each hypothesis
    all_indices
    counter_to_barcode_combo

    """

    def gaussian_updates(
        data: np.ndarray, mu_o: float, std_o: float
    ) -> tuple[float, float]:
        """Update parameters of your gaussian.

        See <https://www.cs.ubc.ca/~murphyk/Papers/bayesGauss.pdf>.

        Parameters
        ----------
        data
            1-d array of counts
        mu_o
            global mean for hashing count distribution
        std_o
            global std for hashing count distribution

        Returns
        -------
        mean
            of gaussian
        std
            of gaussian

        """
        lam_o = 1 / (std_o**2)
        n = len(data)
        lam = 1 / np.var(data) if len(data) > 1 else lam_o
        lam_n = lam_o + n * lam
        mu_n = (
            (np.mean(data) * n * lam + mu_o * lam_o) / lam_n if len(data) > 0 else mu_o
        )
        return mu_n, (1 / (lam_n / (n + 1))) ** (1 / 2)

    eps = 1e-15
    # probabilites for negative, singlet, doublets
    log_likelihoods_for_each_hypothesis = np.zeros((data.shape[0], 3))

    all_indices = np.empty(data.shape[0])
    num_of_barcodes = data.shape[1]
    number_of_non_noise_barcodes = (
        num_of_barcodes - number_of_noise_barcodes
        if number_of_noise_barcodes is not None
        else 2
    )

    num_of_noise_barcodes = num_of_barcodes - number_of_non_noise_barcodes

    # assume log normal
    data = np.log(data + 1)
    data_arg = np.argsort(data, axis=1)
    data_sort = np.sort(data, axis=1)

    # global signal and noise counts useful for when we have few cells
    # barcodes with the highest number of counts are assumed to be a true signal
    # barcodes with rank < k are considered to be noise
    global_signal_counts = np.ravel(data_sort[:, -1])
    global_noise_counts = np.ravel(data_sort[:, :-number_of_non_noise_barcodes])
    global_mu_signal_o, global_sigma_signal_o = (
        np.mean(global_signal_counts),
        np.std(global_signal_counts),
    )
    global_mu_noise_o, global_sigma_noise_o = (
        np.mean(global_noise_counts),
        np.std(global_noise_counts),
    )

    noise_params_dict = {}
    signal_params_dict = {}

    # for each barcode get  empirical noise and signal distribution parameterization
    for x in np.arange(num_of_barcodes):
        sample_barcodes = data[:, x]
        sample_barcodes_noise_idx = np.where(data_arg[:, :num_of_noise_barcodes] == x)[
            0
        ]
        sample_barcodes_signal_idx = np.where(data_arg[:, -1] == x)

        # get noise and signal counts
        noise_counts = sample_barcodes[sample_barcodes_noise_idx]
        signal_counts = sample_barcodes[sample_barcodes_signal_idx]

        # get parameters of distribution, assuming lognormal do update from global values
        noise_param = gaussian_updates(
            noise_counts, global_mu_noise_o, global_sigma_noise_o
        )
        signal_param = gaussian_updates(
            signal_counts, global_mu_signal_o, global_sigma_signal_o
        )
        noise_params_dict[x] = noise_param
        signal_params_dict[x] = signal_param

    counter_to_barcode_combo: dict[int, str] = {}
    counter = 0

    # for each combination of noise and signal barcode calculate probiltiy of in silico and real cell hypotheses
    for noise_sample_idx, signal_sample_idx in product(
        np.arange(num_of_barcodes), np.arange(num_of_barcodes)
    ):
        signal_subset = data_arg[:, -1] == signal_sample_idx
        noise_subset = data_arg[:, -2] == noise_sample_idx
        subset = signal_subset & noise_subset
        if sum(subset) == 0:
            continue

        indices = np.where(subset)[0]
        barcode_combo = "_".join([str(noise_sample_idx), str(signal_sample_idx)])
        all_indices[np.where(subset)[0]] = counter
        counter_to_barcode_combo[counter] = barcode_combo
        counter += 1
        noise_params = noise_params_dict[noise_sample_idx]
        signal_params = signal_params_dict[signal_sample_idx]

        # calculate probabilties for each hypothesis for each cell
        data_subset = data[subset]
        log_signal_signal_probs = np.log(
            norm.pdf(
                data_subset[:, signal_sample_idx],
                *signal_params[:-2],
                loc=signal_params[-2],
                scale=signal_params[-1],
            )
            + eps
        )
        signal_noise_params = signal_params_dict[noise_sample_idx]
        log_noise_signal_probs = np.log(
            norm.pdf(
                data_subset[:, noise_sample_idx],
                loc=signal_noise_params[-2],
                scale=signal_noise_params[-1],
            )
            + eps
        )

        log_noise_noise_probs = np.log(
            norm.pdf(
                data_subset[:, noise_sample_idx],
                loc=noise_params[-2],
                scale=noise_params[-1],
            )
            + eps
        )
        log_signal_noise_probs = np.log(
            norm.pdf(
                data_subset[:, signal_sample_idx],
                loc=noise_params[-2],
                scale=noise_params[-1],
            )
            + eps
        )

        probs_of_negative = np.sum(
            [log_noise_noise_probs, log_signal_noise_probs], axis=0
        )
        probs_of_singlet = np.sum(
            [log_noise_noise_probs, log_signal_signal_probs], axis=0
        )
        probs_of_doublet = np.sum(
            [log_noise_signal_probs, log_signal_signal_probs], axis=0
        )
        log_probs_list = [probs_of_negative, probs_of_singlet, probs_of_doublet]

        # each cell and each hypothesis probability
        for prob_idx, log_prob in enumerate(log_probs_list):
            log_likelihoods_for_each_hypothesis[indices, prob_idx] = log_prob
    return (
        log_likelihoods_for_each_hypothesis,
        all_indices,
        counter_to_barcode_combo,
    )


def _calculate_bayes_rule(
    data: np.ndarray, priors: ArrayLike, number_of_noise_barcodes: int
) -> dict[str, np.ndarray]:
    """Calculate bayes rule from log likelihoods.

    Parameters
    ----------
    data
        Anndata object filled only with hashing counts
    priors
        a list of your prior for each hypothesis
        first element is your prior for the negative hypothesis
        second element is your prior for the singlet hypothesis
        third element is your prior for the doublet hypothesis
        We use [0.01, 0.8, 0.19] by default because we assume the barcodes
        in your cell hashing matrix are those cells which have passed QC
        in the transcriptome space, e.g. UMI counts, pct mito reads, etc.
    number_of_noise_barcodes
        number of barcodes to used to calculated noise distribution

    Returns
    -------
    A dict of bayes key results with the following entries:

    `"most_likely_hypothesis"`
        A 1d np.array of the most likely hypothesis
    `"probs_hypotheses"`
        A 2d np.array probability of each hypothesis
    `"log_likelihoods_for_each_hypothesis"`
        A 2d np.array log likelihood of each hypothesis

    """
    priors = np.array(priors)
    log_likelihoods_for_each_hypothesis, _, _ = _calculate_log_likelihoods(
        data, number_of_noise_barcodes
    )
    probs_hypotheses = (
        np.exp(log_likelihoods_for_each_hypothesis)
        * priors
        / np.sum(
            np.multiply(np.exp(log_likelihoods_for_each_hypothesis), priors),
            axis=1,
        )[:, None]
    )
    most_likely_hypothesis = np.argmax(probs_hypotheses, axis=1)
    return {
        "most_likely_hypothesis": most_likely_hypothesis,
        "probs_hypotheses": probs_hypotheses,
        "log_likelihoods_for_each_hypothesis": log_likelihoods_for_each_hypothesis,
    }


@old_positionals(
    "priors", "pre_existing_clusters", "number_of_noise_barcodes", "inplace"
)
@doctest_skip("Illustrative but not runnable doctest code")
def hashsolo(
    adata: AnnData,
    cell_hashing_columns: Sequence[str],
    *,
    priors: tuple[float, float, float] = (0.01, 0.8, 0.19),
    pre_existing_clusters: str | None = None,
    number_of_noise_barcodes: int | None = None,
    inplace: bool = True,
) -> AnnData | None:
    """Probabilistic demultiplexing of cell hashing data using HashSolo :cite:p:`Bernstein2020`.

    .. note::
        More information and bug reports `here <https://github.com/calico/solo>`__.

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` × `n_vars`.
        Rows correspond to cells and columns to genes.
    cell_hashing_columns
        `.obs` columns that contain cell hashing counts.
    priors
        Prior probabilities of each hypothesis, in
        the order `[negative, singlet, doublet]`. The default is set to
        `[0.01, 0.8, 0.19]` assuming barcode counts are from cells that
        have passed QC in the transcriptome space, e.g. UMI counts, pct
        mito reads, etc.
    pre_existing_clusters
        The column in `.obs` containing pre-existing cluster assignments
        (e.g. Leiden clusters or cell types, but not batch assignments).
        If provided, demultiplexing will be performed separately for each
        cluster.
    number_of_noise_barcodes
        The number of barcodes used to create the noise distribution.
        Defaults to `len(cell_hashing_columns) - 2`.
    inplace
        Whether to update `adata` in-place or return a copy.

    Returns
    -------
    A copy of the input `adata` if `inplace=False`, otherwise the input
    `adata`. The following fields are added:

    `.obs["most_likely_hypothesis"]`
        Index of the most likely hypothesis, where `0` corresponds to negative,
        `1` to singlet, and `2` to doublet.
    `.obs["cluster_feature"]`
        The cluster assignments used for demultiplexing.
    `.obs["negative_hypothesis_probability"]`
        Probability of the negative hypothesis.
    `.obs["singlet_hypothesis_probability"]`
        Probability of the singlet hypothesis.
    `.obs["doublet_hypothesis_probability"]`
        Probability of the doublet hypothesis.
    `.obs["Classification"]`:
        Classification of the cell, one of the barcodes in `cell_hashing_columns`,
        `"Negative"`, or `"Doublet"`.

    Examples
    --------
    >>> import anndata
    >>> import scanpy.external as sce
    >>> adata = anndata.read_h5ad("data.h5ad")
    >>> sce.pp.hashsolo(adata, ["Hash1", "Hash2", "Hash3"])
    >>> adata.obs.head()

    """
    print(
        "Please cite HashSolo paper:\nhttps://www.cell.com/cell-systems/fulltext/S2405-4712(20)30195-2"
    )
    adata = adata.copy() if not inplace else adata
    data = adata.obs[cell_hashing_columns].values
    if not check_nonnegative_integers(data):
        msg = "Cell hashing counts must be non-negative"
        raise ValueError(msg)
    if (number_of_noise_barcodes is not None) and (
        number_of_noise_barcodes >= len(cell_hashing_columns)
    ):
        msg = "number_of_noise_barcodes must be at least one less \
        than the number of samples you have as determined by the number of \
        cell_hashing_columns you've given as input  "
        raise ValueError(msg)
    num_of_cells = adata.shape[0]
    results = pd.DataFrame(
        np.zeros((num_of_cells, 6)),
        columns=[
            "most_likely_hypothesis",
            "probs_hypotheses",
            "cluster_feature",
            "negative_hypothesis_probability",
            "singlet_hypothesis_probability",
            "doublet_hypothesis_probability",
        ],
        index=adata.obs_names,
    )
    if pre_existing_clusters is not None:
        cluster_features = pre_existing_clusters
        unique_cluster_features = np.unique(adata.obs[cluster_features])
        for cluster_feature in unique_cluster_features:
            cluster_feature_bool_vector = adata.obs[cluster_features] == cluster_feature
            posterior_dict = _calculate_bayes_rule(
                data[cluster_feature_bool_vector],
                priors,
                number_of_noise_barcodes,
            )
            results.loc[cluster_feature_bool_vector, "most_likely_hypothesis"] = (
                posterior_dict["most_likely_hypothesis"]
            )
            results.loc[cluster_feature_bool_vector, "cluster_feature"] = (
                cluster_feature
            )
            results.loc[
                cluster_feature_bool_vector, "negative_hypothesis_probability"
            ] = posterior_dict["probs_hypotheses"][:, 0]
            results.loc[
                cluster_feature_bool_vector, "singlet_hypothesis_probability"
            ] = posterior_dict["probs_hypotheses"][:, 1]
            results.loc[
                cluster_feature_bool_vector, "doublet_hypothesis_probability"
            ] = posterior_dict["probs_hypotheses"][:, 2]
    else:
        posterior_dict = _calculate_bayes_rule(data, priors, number_of_noise_barcodes)
        results.loc[:, "most_likely_hypothesis"] = posterior_dict[
            "most_likely_hypothesis"
        ]
        results.loc[:, "cluster_feature"] = 0
        results.loc[:, "negative_hypothesis_probability"] = posterior_dict[
            "probs_hypotheses"
        ][:, 0]
        results.loc[:, "singlet_hypothesis_probability"] = posterior_dict[
            "probs_hypotheses"
        ][:, 1]
        results.loc[:, "doublet_hypothesis_probability"] = posterior_dict[
            "probs_hypotheses"
        ][:, 2]

    adata.obs["most_likely_hypothesis"] = results.loc[
        adata.obs_names, "most_likely_hypothesis"
    ]
    adata.obs["cluster_feature"] = results.loc[adata.obs_names, "cluster_feature"]
    adata.obs["negative_hypothesis_probability"] = results.loc[
        adata.obs_names, "negative_hypothesis_probability"
    ]
    adata.obs["singlet_hypothesis_probability"] = results.loc[
        adata.obs_names, "singlet_hypothesis_probability"
    ]
    adata.obs["doublet_hypothesis_probability"] = results.loc[
        adata.obs_names, "doublet_hypothesis_probability"
    ]

    adata.obs["Classification"] = None
    adata.obs.loc[adata.obs["most_likely_hypothesis"] == 2, "Classification"] = (
        "Doublet"
    )
    adata.obs.loc[adata.obs["most_likely_hypothesis"] == 0, "Classification"] = (
        "Negative"
    )
    all_sings = adata.obs["most_likely_hypothesis"] == 1
    singlet_sample_index = np.argmax(
        adata.obs.loc[all_sings, cell_hashing_columns].values, axis=1
    )
    adata.obs.loc[all_sings, "Classification"] = adata.obs[
        cell_hashing_columns
    ].columns[singlet_sample_index]

    return adata if not inplace else None



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import numpy as np
import pytest
from anndata import AnnData
from packaging.version import Version
from scipy import sparse
from sklearn.neighbors import KNeighborsTransformer

import scanpy as sc
from scanpy import Neighbors
from scanpy._compat import CSBase, pkg_version
from testing.scanpy._helpers import anndata_v0_8_constructor_compat

if TYPE_CHECKING:
    from typing import Literal

    from pytest_mock import MockerFixture

# https://github.com/lmcinnes/umap/issues/1216
SKIPIF_UMAP_BROKEN = pytest.mark.skipif(
    pkg_version("umap-learn") <= Version("0.5.9.post2")
    and pkg_version("numba") >= Version("0.62.0rc1"),
    reason="umap≤0.5.9.post2 is broken with numba≥0.62.0rc1",
)

# the input data
X = [[1, 0], [3, 0], [5, 6], [0, 4]]
n_neighbors = 3  # includes data points themselves

# distances
distances_euclidean = [
    [0.0, 2.0, 0.0, 4.123105525970459],
    [2.0, 0.0, 0.0, 5.0],
    [0.0, 6.324555397033691, 0.0, 5.385164737701416],
    [4.123105525970459, 5.0, 0.0, 0.0],
]

distances_euclidean_all = [
    [0.0, 2.0, 7.211102485656738, 4.123105525970459],
    [2.0, 0.0, 6.324555397033691, 5.0],
    [7.211102485656738, 6.324555397033691, 0.0, 5.385164737701416],
    [4.123105525970459, 5.0, 5.385164737701416, 0.0],
]


# umap "kernel" – only knn results
connectivities_umap = [
    [0.0, 1.0, 0.0, 1.0],
    [1.0, 0.0, 0.5849691143165735, 0.8277419907567016],
    [0.0, 0.5849691143165735, 0.0, 1.0],
    [1.0, 0.8277419907567016, 1.0, 0.0],
]

transitions_sym_umap = [
    [0.0, 0.4818987107873648, 0.0, 0.3951883393150153],
    [0.48189871078736474, 0.0, 0.3594582764005241, 0.24216345431293487],
    [0.0, 0.3594582764005241, 0.0, 0.5039226836320637],
    [0.39518833931501524, 0.24216345431293487, 0.5039226836320637, 0.0],
]

transitions_umap = [
    [0.0, 0.5395987596963403, 0.0, 0.4604012403036599],
    [0.430368608684738, 0.0, 0.3176747629691457, 0.2519566283461165],
    [0.0, 0.40673754271561435, 0.0, 0.5932624572843856],
    [0.33921243006981133, 0.23275092618009624, 0.42803664375009237, 0.0],
]


# gauss kernel [diffmap, dpt] – knn and dense results
connectivities_gauss_knn = [
    [0.0, 0.8466368913650513, 0.0, 0.5660185813903809],
    [0.8466368913650513, 0.0, 0.4223647117614746, 0.4902938902378082],
    [0.0, 0.4223647117614746, 0.0, 0.5840492248535156],
    [0.5660185813903809, 0.4902938902378082, 0.5840492248535156, 0.0],
]

connectivities_gauss_noknn = [
    [1.0, 0.676927387714386, 0.024883469566702843, 0.1962655782699585],
    [0.676927387714386, 1.0, 0.08414449542760849, 0.1353352814912796],
    [0.024883469566702843, 0.08414449542760849, 1.0, 0.16558068990707397],
    [0.1962655782699585, 0.1353352814912796, 0.16558068990707397, 1.0],
]

transitions_sym_gauss_knn = [
    [0.0, 0.5146393179893494, 0.0, 0.36445462703704834],
    [0.5146393179893494, 0.0, 0.3581143319606781, 0.2239987552165985],
    [0.0, 0.3581143319606781, 0.0, 0.5245543718338013],
    [0.36445462703704834, 0.2239987552165985, 0.5245543718338013, 0.0],
]

transitions_sym_gauss_noknn = [
    [
        0.5093212127685547,
        0.34393802285194397,
        0.016115963459014893,
        0.11607448011636734,
    ],
    [0.34393805265426636, 0.506855845451355, 0.054364752024412155, 0.07984541356563568],
    [
        0.016115965321660042,
        0.054364752024412155,
        0.8235670328140259,
        0.12452481687068939,
    ],
    [0.11607448011636734, 0.07984541356563568, 0.1245248094201088, 0.6867417693138123],
]

transitions_gauss_knn = [
    [0.0, 0.5824036598205566, 0.0, 0.4175964295864105],
    [0.4547595679759979, 0.0, 0.3184431493282318, 0.22679725289344788],
    [0.0, 0.4027276933193207, 0.0, 0.5972723364830017],
    [0.3180755078792572, 0.22123482823371887, 0.46068981289863586, 0.0],
]

transitions_gauss_noknn = [
    [0.5093212127685547, 0.3450769782066345, 0.01887294091284275, 0.12672874331474304],
    [0.34280285239219666, 0.506855845451355, 0.06345486640930176, 0.08688655495643616],
    [0.01376173086464405, 0.04657683148980141, 0.8235670328140259, 0.11609435081481934],
    [0.10631592571735382, 0.07337487488985062, 0.13356748223304749, 0.6867417693138123],
]


def get_neighbors() -> Neighbors:
    return Neighbors(anndata_v0_8_constructor_compat(np.array(X)))


@pytest.fixture
def neigh() -> Neighbors:
    return get_neighbors()


@pytest.mark.parametrize("method", ["umap", "gauss"])
def test_distances_euclidean(
    mocker: MockerFixture, neigh: Neighbors, method: Literal["umap", "gauss"]
):
    """Umap and gauss behave the same for distances.

    They call pynndescent for large data.
    """
    from pynndescent import NNDescent

    # When trying to compress a too-small index, pynndescent complains
    mocker.patch.object(NNDescent, "compress_index", return_val=None)

    neigh.compute_neighbors(n_neighbors, method=method)
    np.testing.assert_allclose(neigh.distances.toarray(), distances_euclidean)


@pytest.mark.parametrize(
    ("transformer", "knn"),
    [
        # knn=False trivially returns all distances
        pytest.param(None, False, id="knn=False"),
        # pynndescent returns all distances when data is so small
        pytest.param("pynndescent", True, id="pynndescent"),
        # Explicit brute force also returns all distances
        pytest.param(
            KNeighborsTransformer(n_neighbors=n_neighbors, algorithm="brute"),
            True,
            id="sklearn",
        ),
    ],
)
def test_distances_all(neigh: Neighbors, transformer, knn):
    neigh.compute_neighbors(
        n_neighbors, transformer=transformer, method="gauss", knn=knn
    )
    dists = (
        neigh.distances.toarray()
        if isinstance(neigh.distances, CSBase)
        else neigh.distances
    )
    np.testing.assert_allclose(dists, distances_euclidean_all)


@pytest.mark.parametrize(
    ("method", "conn", "trans", "trans_sym"),
    [
        pytest.param(
            "umap",
            connectivities_umap,
            transitions_umap,
            transitions_sym_umap,
            marks=SKIPIF_UMAP_BROKEN,
            id="umap",
        ),
        pytest.param(
            "gauss",
            connectivities_gauss_knn,
            transitions_gauss_knn,
            transitions_sym_gauss_knn,
            id="gauss",
        ),
    ],
)
def test_connectivities_euclidean(neigh: Neighbors, method, conn, trans, trans_sym):
    neigh.compute_neighbors(n_neighbors, method=method)
    np.testing.assert_allclose(neigh.connectivities.toarray(), conn)
    neigh.compute_transitions()
    np.testing.assert_allclose(neigh.transitions_sym.toarray(), trans_sym, rtol=1e-5)
    np.testing.assert_allclose(neigh.transitions.toarray(), trans, rtol=1e-5)


def test_gauss_noknn_connectivities_euclidean(neigh):
    neigh.compute_neighbors(n_neighbors, method="gauss", knn=False)
    np.testing.assert_allclose(neigh.connectivities, connectivities_gauss_noknn)
    neigh.compute_transitions()
    np.testing.assert_allclose(
        neigh.transitions_sym, transitions_sym_gauss_noknn, rtol=1e-5
    )
    np.testing.assert_allclose(neigh.transitions, transitions_gauss_noknn, rtol=1e-5)


def test_metrics_argument():
    no_knn_euclidean = get_neighbors()
    no_knn_euclidean.compute_neighbors(
        n_neighbors, method="gauss", knn=False, metric="euclidean"
    )
    no_knn_manhattan = get_neighbors()
    no_knn_manhattan.compute_neighbors(
        n_neighbors, method="gauss", knn=False, metric="manhattan"
    )
    assert not np.allclose(no_knn_euclidean.distances, no_knn_manhattan.distances)


def test_use_rep_argument():
    adata = AnnData(np.random.randn(30, 300))
    sc.pp.pca(adata)
    neigh_pca = Neighbors(adata)
    neigh_pca.compute_neighbors(n_pcs=5, use_rep="X_pca")
    neigh_none = Neighbors(adata)
    neigh_none.compute_neighbors(n_pcs=5, use_rep=None)
    np.testing.assert_allclose(
        neigh_pca.distances.toarray(), neigh_none.distances.toarray()
    )


@pytest.mark.parametrize("conv", [sparse.csr_matrix.toarray, sparse.csr_matrix])  # noqa: TID251
def test_restore_n_neighbors(neigh, conv):
    neigh.compute_neighbors(n_neighbors, method="gauss")

    ad = AnnData(np.array(X))
    # Allow deprecated usage for now
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=FutureWarning, module="anndata")
        ad.uns["neighbors"] = dict(connectivities=conv(neigh.connectivities))
    neigh_restored = Neighbors(ad)
    assert neigh_restored.n_neighbors == 1



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pytest
from sklearn.neighbors import KNeighborsTransformer

from scanpy._utils.compute.is_constant import is_constant
from scanpy.neighbors._common import (
    _get_sparse_matrix_from_indices_distances,
    _has_self_column,
    _ind_dist_shortcut,
)

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Literal

    from scipy import sparse

    from scanpy._compat import CSRBase


def mk_knn_matrix(
    n_obs: int,
    n_neighbors: int,
    *,
    style: Literal["basic", "rapids", "sklearn"],
    duplicates: bool = False,
) -> CSRBase:
    n_col = n_neighbors + (1 if style == "sklearn" else 0)
    dists = np.abs(np.random.randn(n_obs, n_col)) + 1e-8
    idxs = np.arange(n_obs * n_col).reshape((n_col, n_obs)).T
    if style == "rapids":
        idxs[:, 0] += 1  # does not include cell itself
    else:
        dists[:, 0] = 0.0  # includes cell itself
    if duplicates:
        # Don’t use the first column, as that might be the cell itself
        dists[n_obs // 4 : n_obs, 2] = 0.0
    # keep self column to simulate output from kNN transformers
    mat = _get_sparse_matrix_from_indices_distances(idxs, dists, keep_self=True)

    # check if out helper here works as expected
    assert _has_self_column(idxs, dists) == (style != "rapids")
    if duplicates:
        # Make sure the actual matrix has a regular sparsity pattern
        assert is_constant(mat.getnnz(axis=1))
        # Make sure implicit zeros for duplicates would change the sparsity pattern
        mat_sparsified = mat.copy()
        mat_sparsified.eliminate_zeros()
        assert not is_constant(mat_sparsified.getnnz(axis=1))

    return mat


@pytest.mark.parametrize("n_neighbors", [3, pytest.param(None, id="all")])
@pytest.mark.parametrize("style", ["basic", "rapids", "sklearn"])
@pytest.mark.parametrize("duplicates", [True, False], ids=["duplicates", "unique"])
def test_ind_dist_shortcut_manual(
    *,
    n_neighbors: int | None,
    style: Literal["basic", "rapids", "sklearn"],
    duplicates: bool,
):
    n_obs = 10
    if n_neighbors is None:
        n_neighbors = n_obs
    mat = mk_knn_matrix(n_obs, n_neighbors, style=style, duplicates=duplicates)

    assert (mat.nnz / n_obs) == n_neighbors + (1 if style == "sklearn" else 0)
    assert _ind_dist_shortcut(mat) is not None


@pytest.mark.parametrize("n_neighbors", [3, pytest.param(None, id="all")])
@pytest.mark.parametrize(
    "mk_mat",
    [
        pytest.param(
            lambda n_obs, n_neighbors: KNeighborsTransformer(
                n_neighbors=n_neighbors
            ).fit_transform(np.random.randn(n_obs, n_obs // 4)),
            id="sklearn_auto",
        )
    ],
)
def test_ind_dist_shortcut_premade(
    n_neighbors: int | None,
    mk_mat: Callable[[int, int], sparse.csr_matrix],  # noqa: TID251
):
    n_obs = 10
    if n_neighbors is None:
        # KNeighborsTransformer interprets this as “number of neighbors excluding cell itself”
        # so it can be at most n_obs - 1
        n_neighbors = n_obs - 1
    mat = mk_mat(n_obs, n_neighbors)

    assert (mat.nnz / n_obs) == n_neighbors + 1
    assert _ind_dist_shortcut(mat) is not None



"""Run Diffusion maps using the adaptive anisotropic kernel."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd

from ... import logging as logg
from ..._compat import old_positionals
from ..._utils._doctests import doctest_needs

if TYPE_CHECKING:
    from anndata import AnnData


@old_positionals(
    "n_components",
    "knn",
    "alpha",
    "use_adjacency_matrix",
    "distances_key",
    "n_eigs",
    "impute_data",
    "n_steps",
    "copy",
)
@doctest_needs("palantir")
def palantir(
    adata: AnnData,
    *,
    n_components: int = 10,
    knn: int = 30,
    alpha: float = 0,
    use_adjacency_matrix: bool = False,
    distances_key: str | None = None,
    n_eigs: int | None = None,
    impute_data: bool = True,
    n_steps: int = 3,
    copy: bool = False,
) -> AnnData | None:
    """Run Diffusion maps using the adaptive anisotropic kernel :cite:p:`Setty2019`.

    Palantir is an algorithm to align cells along differentiation trajectories.
    Palantir models differentiation as a stochastic process where stem cells
    differentiate to terminally differentiated cells by a series of steps through
    a low dimensional phenotypic manifold. Palantir effectively captures the
    continuity in cell states and the stochasticity in cell fate determination.
    Palantir has been designed to work with multidimensional single cell data
    from diverse technologies such as Mass cytometry and single cell RNA-seq.

    .. note::
       More information and bug reports `here <https://github.com/dpeerlab/Palantir>`__.

    Parameters
    ----------
    adata
        An AnnData object.
    n_components
        Number of diffusion components.
    knn
        Number of nearest neighbors for graph construction.
    alpha
        Normalization parameter for the diffusion operator.
    use_adjacency_matrix
        Use adaptive anisotropic adjacency matrix, instead of PCA projections
        (default) to compute diffusion components.
    distances_key
        With `use_adjacency_matrix=True`, use the indicated distances key for `.obsp`.
        If `None`, `'distances'`.
    n_eigs
        Number of eigen vectors to use. If `None` specified, the number of eigen
        vectors will be determined using eigen gap. Passed to
        `palantir.utils.determine_multiscale_space`.
    impute_data
        Impute data using MAGIC.
    n_steps
        Number of steps in the diffusion operator. Passed to
        `palantir.utils.run_magic_imputation`.
    copy
        Return a copy instead of writing to `adata`.

    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields:

    **Diffusion maps**,
        used for magic imputation, and to generate multi-scale data matrix,

        - X_palantir_diff_comp - :class:`~numpy.ndarray` (:attr:`~anndata.AnnData.obsm`, dtype `float`)
            Array of Diffusion components.
        - palantir_EigenValues - :class:`~numpy.ndarray` (:attr:`~anndata.AnnData.uns`, dtype `float`)
            Array of corresponding eigen values.
        - palantir_diff_op - :class:`~scipy.sparse.csr_matrix` (:attr:`~anndata.AnnData.obsp`, dtype `float`)
            The diffusion operator matrix.

    **Multi scale space results**,
        used to build tsne on diffusion components, and to compute branch probabilities
        and waypoints,

        - X_palantir_multiscale - :class:`~numpy.ndarray` (:attr:`~anndata.AnnData.obsm`, dtype `float`)
            Multi scale data matrix.

    **MAGIC imputation**,
        used for plotting gene expression on tsne, and gene expression trends,

        - palantir_imp - :class:`~numpy.ndarray` (:attr:`~anndata.AnnData.layers`, dtype `float`)
            Imputed data matrix (MAGIC imputation).

    Example
    -------
    >>> import scanpy.external as sce
    >>> import scanpy as sc

    A sample data is available `here <https://github.com/dpeerlab/Palantir/tree/master/data>`_.

    **Load sample data**

    >>> adata = sc.read_csv(filename="Palantir/data/marrow_sample_scseq_counts.csv.gz")

    *Cleanup and normalize*

    >>> sc.pp.filter_cells(adata, min_counts=1000)
    >>> sc.pp.filter_genes(adata, min_counts=10)
    >>> sc.pp.normalize_per_cell(adata)
    >>> sc.pp.log1p(adata)

    **Data preprocessing**

    Palantir builds diffusion maps using one of two optional inputs:

    *Principal component analysis*

    >>> sc.pp.pca(adata, n_comps=300)

    or,

    *Nearist neighbors graph*

    >>> sc.pp.neighbors(adata, knn=30)

    *Diffusion maps*

    Palantir determines the diffusion maps of the data as an estimate of the low
    dimensional phenotypic manifold of the data.

    >>> sce.tl.palantir(adata, n_components=5, knn=30)

    if pre-computed distances are to be used,

    >>> sce.tl.palantir(
    ...     adata,
    ...     n_components=5,
    ...     knn=30,
    ...     use_adjacency_matrix=True,
    ...     distances_key="distances",
    ... )

    **Visualizing Palantir results**

    *tSNE visualization*

    important for Palantir!

    Palantir constructs the tSNE map in the embedded space since these maps better
    represent the differentiation trajectories.

    >>> sc.tl.tsne(adata, n_pcs=2, use_rep="X_palantir_multiscale", perplexity=150)

    *tsne by cell size*

    >>> sc.pl.tsne(adata, color="n_counts")

    *Imputed gene expression visualized on tSNE maps*

    >>> sc.pl.tsne(
    ...     adata,
    ...     gene_symbols=["CD34", "MPO", "GATA1", "IRF8"],
    ...     layer="palantir_imp",
    ...     color=["CD34", "MPO", "GATA1", "IRF8"],
    ... )

    **Running Palantir**

    Palantir can be run by specifying an approximate early cell. While Palantir
    automatically determines the terminal states, they can also be specified using the
    `termine_states` parameter.

    >>> start_cell = "Run5_164698952452459"
    >>> pr_res = sce.tl.palantir_results(
    ...     adata,
    ...     early_cell=start_cell,
    ...     ms_data="X_palantir_multiscale",
    ...     num_waypoints=500,
    ... )

    .. note::
       A `start_cell` must be defined for every data set. The start cell for
       this dataset was chosen based on high expression of CD34.

    At this point the returned Palantir object `pr_res` can be used for all downstream
    analysis and plotting. Please consult this notebook
    `Palantir_sample_notebook.ipynb
    <https://github.com/dpeerlab/Palantir/blob/master/notebooks/Palantir_sample_notebook.ipynb>`_.
    It provides a comprehensive guide to draw *gene expression trends*, amongst other
    things.

    """
    _check_import()
    from palantir.utils import (
        determine_multiscale_space,
        run_diffusion_maps,
        run_magic_imputation,
    )

    adata = adata.copy() if copy else adata

    logg.info("Palantir Diffusion Maps in progress ...")

    if use_adjacency_matrix:
        df = adata.obsp[distances_key] if distances_key else adata.obsp["distances"]
    else:
        df = pd.DataFrame(adata.obsm["X_pca"], index=adata.obs_names)

    # Diffusion maps
    dm_res = run_diffusion_maps(
        df,
        n_components=n_components,
        knn=knn,
        alpha=alpha,
    )
    # Determine the multi scale space of the data
    ms_data = determine_multiscale_space(dm_res=dm_res, n_eigs=n_eigs)

    # MAGIC imputation
    if impute_data:
        imp_df = run_magic_imputation(
            data=adata.to_df(), dm_res=dm_res, n_steps=n_steps
        )
        adata.layers["palantir_imp"] = imp_df

    (
        adata.obsm["X_palantir_diff_comp"],
        adata.uns["palantir_EigenValues"],
        adata.obsp["palantir_diff_op"],
        adata.obsm["X_palantir_multiscale"],
    ) = (
        dm_res["EigenVectors"].to_numpy(),
        dm_res["EigenValues"].to_numpy(),
        dm_res["T"],
        ms_data.to_numpy(),
    )

    return adata if copy else None


@old_positionals(
    "ms_data",
    "terminal_states",
    "knn",
    "num_waypoints",
    "n_jobs",
    "scale_components",
    "use_early_cell_as_start",
    "max_iterations",
)
def palantir_results(
    adata: AnnData,
    early_cell: str,
    *,
    ms_data: str = "X_palantir_multiscale",
    terminal_states: list | None = None,
    knn: int = 30,
    num_waypoints: int = 1200,
    n_jobs: int = -1,
    scale_components: bool = True,
    use_early_cell_as_start: bool = False,
    max_iterations: int = 25,
) -> AnnData | None:
    """Run Palantir.

    A convenience function that wraps `palantir.core.run_palantir` to compute branch
    probabilities and waypoints.

    Parameters
    ----------
    adata
        An AnnData object.
    early_cell
        Start cell for pseudotime construction.
    ms_data
        Palantir multi scale data matrix,
    terminal_states
        List of user defined terminal states
    knn
        Number of nearest neighbors for graph construction.
    num_waypoints
        Number of waypoints to sample.
    n_jobs
        Number of jobs for parallel processing.
    scale_components
        Transform features by scaling each feature to a given range. Consult the
        documentation for `sklearn.preprocessing.minmax_scale`.
    use_early_cell_as_start
        Use `early_cell` as `start_cell`, instead of determining it from the boundary
        cells closest to the defined `early_cell`.
    max_iterations
        Maximum number of iterations for pseudotime convergence.

    Returns
    -------
    PResults object with pseudotime, entropy, branch probabilities and waypoints.

    """
    logg.info("Palantir computing waypoints..")

    _check_import()
    from palantir.core import run_palantir

    ms_data = pd.DataFrame(adata.obsm[ms_data], index=adata.obs_names)
    pr_res = run_palantir(
        ms_data,
        early_cell=early_cell,
        terminal_states=terminal_states,
        knn=knn,
        num_waypoints=num_waypoints,
        n_jobs=n_jobs,
        scale_components=scale_components,
        use_early_cell_as_start=use_early_cell_as_start,
        max_iterations=max_iterations,
    )

    return pr_res


def _check_import():
    try:
        import palantir  # noqa: F401
    except ImportError as e:
        msg = "\nplease install palantir:\n\tpip install palantir"
        raise ImportError(msg) from e



"""Helper functions for the scanpy test suite."""

from __future__ import annotations

import warnings
from contextlib import AbstractContextManager, contextmanager
from dataclasses import dataclass
from importlib.util import find_spec
from itertools import permutations
from types import MappingProxyType
from typing import TYPE_CHECKING

import numpy as np
from anndata.tests.helpers import asarray, assert_equal

import scanpy as sc
from scanpy._compat import pkg_version

if TYPE_CHECKING:
    from collections.abc import Iterable, MutableSequence

    from numpy.typing import NDArray

    from scanpy._compat import DaskArray

# TODO: Report more context on the fields being compared on error
# TODO: Allow specifying paths to ignore on comparison

###########################
# Representation choice
###########################
# These functions can be used to check that functions are correctly using arugments like `layers`, `obsm`, etc.


def anndata_v0_8_constructor_compat(x, /, *args, **kwargs):
    """Construct AnnData that uses dtype of X for test compatibility with older AnnData versions.

    Once the minimum version of AnnData is 0.9, this function can be replaced with the default constructor.
    """
    import anndata as ad
    from packaging.version import Version

    if pkg_version("anndata") < Version("0.9"):
        return ad.AnnData(x, *args, **kwargs, dtype=x.dtype)
    else:
        return ad.AnnData(x, *args, **kwargs)


def check_rep_mutation(func, x, *, fields=("layer", "obsm"), **kwargs):
    """Check that only the array meant to be modified is modified."""
    adata_in = anndata_v0_8_constructor_compat(x.copy())

    for field in fields:
        sc.get._set_obs_rep(adata_in, x, **{field: field})
    x_array = asarray(x)

    adata_out = func(adata_in, copy=True, **kwargs)
    adatas_proc = {
        field: func(adata_in, copy=True, **{field: field}, **kwargs) for field in fields
    }

    # Modified fields
    for field in fields:
        result_array = asarray(
            sc.get._get_obs_rep(adatas_proc[field], **{field: field})
        )
        np.testing.assert_array_equal(asarray(adata_out.X), result_array)

    # Unmodified fields
    for field in fields:
        np.testing.assert_array_equal(x_array, asarray(adatas_proc[field].X))
        np.testing.assert_array_equal(
            x_array, asarray(sc.get._get_obs_rep(adata_out, **{field: field}))
        )
    for field_a, field_b in permutations(fields, 2):
        result_array = asarray(
            sc.get._get_obs_rep(adatas_proc[field_a], **{field_b: field_b})
        )
        np.testing.assert_array_equal(x_array, result_array)


def check_rep_results(func, x, *, fields: Iterable[str] = ("layer", "obsm"), **kwargs):
    """Check that the results of a computation add values/ mutate the anndata object in a consistent way."""
    # Gen data
    empty_x = np.zeros(shape=x.shape, dtype=x.dtype)
    adata_empty = sc.AnnData(
        X=empty_x.copy(),
        layers={"layer": empty_x.copy()},
        obsm={"obsm": empty_x.copy()},
    )

    adata = adata_empty.copy()
    adata.X = x.copy()

    adatas_proc = {}
    for field in fields:
        cur = adata_empty.copy()
        sc.get._set_obs_rep(cur, x.copy(), **{field: field})
        adatas_proc[field] = cur

    # Apply function
    func(adata, **kwargs)
    for field in fields:
        func(adatas_proc[field], **{field: field}, **kwargs)
    # Reset X
    adata.X = empty_x.copy()
    for field in fields:
        sc.get._set_obs_rep(adatas_proc[field], empty_x.copy(), **{field: field})

    for field_a, field_b in permutations(fields, 2):
        assert_equal(adatas_proc[field_a], adatas_proc[field_b])
    for field in fields:
        assert_equal(adata, adatas_proc[field])


def _check_check_values_warnings(
    function, adata, expected_warning, kwargs=MappingProxyType({})
):
    """Run `function` on `adata` with provided arguments `kwargs` twice.

    Once with `check_values=True` and once with `check_values=False`.
    Checks that the `expected_warning` is only raised whtn `check_values=True`.
    """
    # expecting 0 no-int warnings
    with warnings.catch_warnings(record=True) as record:
        function(adata.copy(), **kwargs, check_values=False)
    warning_msgs = [w.message.args[0] for w in record]
    assert expected_warning not in warning_msgs

    # expecting 1 no-int warning
    with warnings.catch_warnings(record=True) as record:
        function(adata.copy(), **kwargs, check_values=True)
    warning_msgs = [w.message.args[0] for w in record]
    assert expected_warning in warning_msgs


# Delayed imports for case where we aren't using dask
def as_dense_dask_array(*args, **kwargs) -> DaskArray:
    from anndata.tests.helpers import as_dense_dask_array

    return as_dense_dask_array(*args, **kwargs)


def as_sparse_dask_array(*args, **kwargs) -> DaskArray:
    from anndata.tests.helpers import as_sparse_dask_array

    return as_sparse_dask_array(*args, **kwargs)


@dataclass(init=False)
class MultiContext(AbstractContextManager):
    contexts: MutableSequence[AbstractContextManager]

    def __init__(self, *contexts: AbstractContextManager):
        self.contexts = list(contexts)

    def __enter__(self):
        for ctx in self.contexts:
            ctx.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        for ctx in reversed(self.contexts):
            ctx.__exit__(exc_type, exc_value, traceback)


@contextmanager
def maybe_dask_process_context():
    """Switch to a single-threaded scheduler for tests that use numba.

    Running numba with dask's threaded scheduler causes crashes,
    so we need to switch to single-threaded (or processes, which is slower).
    """
    if not find_spec("dask"):
        yield
        return

    import dask.config

    prev_scheduler = dask.config.get("scheduler", "threads")
    dask.config.set(scheduler="single-threaded")
    try:
        yield
    finally:
        dask.config.set(scheduler=prev_scheduler)


def random_mask(n: int, *, rng: np.random.Generator | None = None) -> NDArray[np.bool_]:
    """Generate a random mask.

    Makes sure that at least 2 mask entries are True and at least 2 are False.
    This avoids off-by-1 errors even in e.g. neighbors (which already cuts 1 off).
    """
    assert n >= 4, "n must be at least 4"
    rng = np.random.default_rng(rng)
    mask = rng.choice([True, False], n)
    if (n_false := (~mask).sum()) < 2:
        mask[rng.choice(np.flatnonzero(mask), 2 - n_false, replace=False)] = False
    if (n_true := mask.sum()) < 2:
        mask[rng.choice(np.flatnonzero(~mask), 2 - n_true, replace=False)] = True
    return mask



"""Tests to make sure the example datasets load."""

from __future__ import annotations

import subprocess
import warnings
from collections import defaultdict
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import numpy as np
import pytest
from anndata.tests.helpers import assert_adata_equal

import scanpy as sc
from testing.scanpy._helpers import data
from testing.scanpy._pytest.marks import needs

if TYPE_CHECKING:
    from collections.abc import Callable

    from anndata import AnnData


@pytest.fixture(autouse=True)
def _tmp_dataset_dir(tmp_path: Path) -> None:
    """Make sure that datasets are downloaded during the test run.

    The default test environment stores them in a cached location.
    """
    sc.settings.datasetdir = tmp_path / "scanpy_data"


@pytest.mark.internet
def test_burczynski06():
    with pytest.warns(UserWarning, match=r"Variable names are not unique"):
        adata = sc.datasets.burczynski06()
    assert adata.shape == (127, 22283)
    assert not (adata.X == 0).any()


@pytest.mark.internet
@needs.openpyxl
def test_moignard15():
    with warnings.catch_warnings():
        # https://foss.heptapod.net/openpyxl/openpyxl/-/issues/2051
        warnings.filterwarnings(
            "ignore",
            r"datetime\.datetime\.utcnow\(\) is deprecated",
            category=DeprecationWarning,
            module="openpyxl",
        )
        adata = sc.datasets.moignard15()
    assert adata.shape == (3934, 42)


@pytest.mark.internet
def test_paul15():
    sc.datasets.paul15()


@pytest.mark.internet
def test_pbmc3k():
    adata = sc.datasets.pbmc3k()
    assert adata.shape == (2700, 32738)
    assert "CD8A" in adata.var_names


@pytest.mark.internet
def test_pbmc3k_processed():
    with warnings.catch_warnings(record=True) as records:
        adata = sc.datasets.pbmc3k_processed()
    assert adata.shape == (2638, 1838)
    assert adata.raw.shape == (2638, 13714)

    assert len(records) == 0


@pytest.mark.internet
def test_ebi_expression_atlas(monkeypatch: pytest.MonkeyPatch):
    from scanpy.datasets import _ebi_expression_atlas as ea_mod

    # make sure we use chunks when testing.
    # This dataset has <8M entries, so 4M entries/chunk = 2 chunks
    assert hasattr(ea_mod, "CHUNK_SIZE")
    monkeypatch.setattr(ea_mod, "CHUNK_SIZE", int(4e6))

    adata = sc.datasets.ebi_expression_atlas("E-MTAB-4888")
    # The shape changes sometimes
    assert 2261 <= adata.shape[0] <= 2315
    assert 23899 <= adata.shape[1] <= 24051


def test_krumsiek11():
    with pytest.warns(UserWarning, match=r"Observation names are not unique"):
        adata = sc.datasets.krumsiek11()
    assert adata.shape == (640, 11)
    assert set(adata.obs["cell_type"]) == {"Ery", "Mk", "Mo", "Neu", "progenitor"}


def test_blobs():
    n_obs = np.random.randint(15, 30)
    n_var = np.random.randint(500, 600)
    adata = sc.datasets.blobs(n_variables=n_var, n_observations=n_obs)
    assert adata.shape == (n_obs, n_var)


def test_toggleswitch():
    with pytest.warns(UserWarning, match=r"Observation names are not unique"):
        sc.datasets.toggleswitch()


def test_pbmc68k_reduced():
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        sc.datasets.pbmc68k_reduced()


@pytest.mark.filterwarnings("ignore:Use `squidpy.*` instead:FutureWarning")
@pytest.mark.internet
def test_visium_datasets():
    """Tests that reading/ downloading works and is does not have global effects."""
    with pytest.warns(UserWarning, match=r"Variable names are not unique"):
        hheart = sc.datasets.visium_sge("V1_Human_Heart")
    with pytest.warns(UserWarning, match=r"Variable names are not unique"):
        hheart_again = sc.datasets.visium_sge("V1_Human_Heart")
    assert_adata_equal(hheart, hheart_again)


@pytest.mark.filterwarnings("ignore:Use `squidpy.*` instead:FutureWarning")
@pytest.mark.internet
def test_visium_datasets_dir_change(tmp_path: Path):
    """Test that changing the dataset dir doesn't break reading."""
    with pytest.warns(UserWarning, match=r"Variable names are not unique"):
        mbrain = sc.datasets.visium_sge("V1_Adult_Mouse_Brain")
    sc.settings.datasetdir = tmp_path
    with pytest.warns(UserWarning, match=r"Variable names are not unique"):
        mbrain_again = sc.datasets.visium_sge("V1_Adult_Mouse_Brain")
    assert_adata_equal(mbrain, mbrain_again)


@pytest.mark.filterwarnings("ignore:Use `squidpy.*` instead:FutureWarning")
@pytest.mark.internet
def test_visium_datasets_images():
    """Test that image download works and is does not have global effects."""
    # Test that downloading tissue image works
    with pytest.warns(UserWarning, match=r"Variable names are not unique"):
        mbrain = sc.datasets.visium_sge("V1_Adult_Mouse_Brain", include_hires_tiff=True)
    expected_image_path = sc.settings.datasetdir / "V1_Adult_Mouse_Brain" / "image.tif"
    image_path = Path(
        mbrain.uns["spatial"]["V1_Adult_Mouse_Brain"]["metadata"]["source_image_path"]
    )
    assert image_path == expected_image_path

    # Test that tissue image exists and is a valid image file
    assert image_path.exists()

    # Test that tissue image is a tif image file (using `file`)
    process = subprocess.run(
        ["file", "--mime-type", image_path], stdout=subprocess.PIPE, check=True
    )
    output = process.stdout.strip().decode()  # make process output string
    assert output == f"{image_path}: image/tiff"


def test_download_failure():
    from urllib.error import HTTPError

    with pytest.raises(HTTPError):
        sc.datasets.ebi_expression_atlas("not_a_real_accession")


# These are tested via doctest
DS_INCLUDED = frozenset({"krumsiek11", "toggleswitch", "pbmc68k_reduced"})
# These have parameters that affect shape and so on
DS_DYNAMIC = frozenset({"ebi_expression_atlas"})
# Additional marks for datasets besides “internet”
DS_MARKS = defaultdict(list, moignard15=[needs.openpyxl])


@pytest.mark.parametrize(
    "ds_name",
    [
        pytest.param(
            ds,
            id=ds,
            marks=[
                *(() if ds in DS_INCLUDED else [pytest.mark.internet]),
                *DS_MARKS[ds],
            ],
        )
        for ds in sorted(set(sc.datasets.__all__) - DS_DYNAMIC)
    ],
)
def test_doc_shape(ds_name):
    dataset_fn: Callable[[], AnnData] = getattr(sc.datasets, ds_name)
    assert dataset_fn.__doc__, "No docstring"
    start_line_2 = dataset_fn.__doc__.find("\n") + 1
    docstring = dedent(dataset_fn.__doc__[start_line_2:])
    cached_fn = getattr(data, ds_name, dataset_fn)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            r"(Observation|Variable) names are not unique",
            category=UserWarning,
        )
        dataset = cached_fn()
    assert repr(dataset) in docstring



"""Functions returning copies of datasets as cheaply as possible.

i.e. without having to hit the disk or (in case of ``_pbmc3k_normalized``) recomputing normalization.
"""

from __future__ import annotations

import warnings
from functools import cache
from typing import TYPE_CHECKING

import scanpy as sc

if TYPE_CHECKING:
    from anndata import AnnData

# Functions returning the same objects (easy to misuse)


_pbmc3k = cache(sc.datasets.pbmc3k)
_pbmc3k_processed = cache(sc.datasets.pbmc3k_processed)
_pbmc68k_reduced = cache(sc.datasets.pbmc68k_reduced)
_krumsiek11 = cache(sc.datasets.krumsiek11)
_paul15 = cache(sc.datasets.paul15)


# Functions returning copies


def pbmc3k() -> AnnData:
    return _pbmc3k().copy()


def pbmc3k_processed() -> AnnData:
    return _pbmc3k_processed().copy()


def pbmc68k_reduced() -> AnnData:
    return _pbmc68k_reduced().copy()


def krumsiek11() -> AnnData:
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", "Observation names are not unique", module="anndata"
        )
        return _krumsiek11().copy()


def paul15() -> AnnData:
    return _paul15().copy()


# Derived datasets


@cache
def _pbmc3k_normalized() -> AnnData:
    pbmc = pbmc3k()
    pbmc.X = pbmc.X.astype("float64")  # For better accuracy
    sc.pp.filter_genes(pbmc, min_counts=1)
    sc.pp.log1p(pbmc)
    sc.pp.normalize_total(pbmc)
    sc.pp.highly_variable_genes(pbmc)
    return pbmc


def pbmc3k_normalized() -> AnnData:
    return _pbmc3k_normalized().copy()



from __future__ import annotations

import scanpy.external as sce
from testing.scanpy._helpers.data import pbmc3k_processed
from testing.scanpy._pytest.marks import needs

pytestmark = [needs.palantir]


def test_palantir_core():
    adata = pbmc3k_processed()

    sce.tl.palantir(adata=adata, n_components=5, knn=30)
    assert adata.layers["palantir_imp"].shape[0], "palantir_imp matrix Error!"



from __future__ import annotations

import pytest

import scanpy as sc
from testing.scanpy._helpers.data import pbmc68k_reduced


def test_deprecate_multicore_tsne():
    pbmc = pbmc68k_reduced()

    with pytest.warns(
        UserWarning, match="calling tsne with n_jobs > 1 would use MulticoreTSNE"
    ):
        sc.tl.tsne(pbmc, n_jobs=2)

    with pytest.warns(FutureWarning, match="Argument `use_fast_tsne` is deprecated"):
        sc.tl.tsne(pbmc, use_fast_tsne=True)

    with pytest.warns(UserWarning, match="Falling back to scikit-learn"):
        sc.tl.tsne(pbmc, use_fast_tsne=True)


def test_deprecate_use_highly_variable_genes():
    pbmc = pbmc68k_reduced()

    with pytest.warns(
        FutureWarning, match="Argument `use_highly_variable` is deprecated"
    ):
        sc.pp.pca(pbmc, use_highly_variable=True)



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Literal, Protocol

import numpy as np

if TYPE_CHECKING:
    from typing import Any, Self

    from .._compat import CSRBase


# These two are used with get_literal_vals elsewhere
_Method = Literal["umap", "gauss"]
_KnownTransformer = Literal["pynndescent", "sklearn", "rapids"]

# sphinx-autodoc-typehints can’t transitively import types from if TYPE_CHECKING blocks,
# so these four needs to be importable

_MetricFn = Callable[[np.ndarray, np.ndarray], float]
# from sklearn.metrics.pairwise_distances.__doc__:
_MetricSparseCapable = Literal[
    "cityblock", "cosine", "euclidean", "l1", "l2", "manhattan"
]
_MetricScipySpatial = Literal[
    "braycurtis",
    "canberra",
    "chebyshev",
    "correlation",
    "dice",
    "hamming",
    "jaccard",
    "kulsinski",
    "mahalanobis",
    "minkowski",
    "rogerstanimoto",
    "russellrao",
    "seuclidean",
    "sokalmichener",
    "sokalsneath",
    "sqeuclidean",
    "yule",
]
_Metric = _MetricSparseCapable | _MetricScipySpatial


class KnnTransformerLike(Protocol):
    """See :class:`~sklearn.neighbors.KNeighborsTransformer`."""

    def fit(self, x, /, y: None = None): ...
    def transform(self, x, /) -> CSRBase: ...

    # from TransformerMixin
    def fit_transform(self, x, /, y: None = None) -> CSRBase: ...

    # from BaseEstimator
    def get_params(self, *, deep: bool = True) -> dict[str, Any]: ...
    def set_params(self, **params: Any) -> Self: ...



"""Moran's I global spatial autocorrelation."""

from __future__ import annotations

from functools import singledispatch
from typing import TYPE_CHECKING, cast

import numba
import numpy as np

from .._compat import CSRBase, _register_union, njit
from ..get import _get_obs_rep
from ._common import _get_graph, _SparseMetric

if TYPE_CHECKING:
    from anndata import AnnData
    from numpy.typing import NDArray

    from ._common import _Vals


@singledispatch
def morans_i(
    adata_or_graph: AnnData | CSRBase,
    /,
    vals: _Vals | None = None,
    *,
    use_graph: str | None = None,
    layer: str | None = None,
    obsm: str | None = None,
    obsp: str | None = None,
    use_raw: bool = False,
) -> np.ndarray | float:
    r"""Calculate Moran’s I Global Autocorrelation Statistic.

    Moran’s I is a global autocorrelation statistic for some measure on a graph. It is commonly used in
    spatial data analysis to assess autocorrelation on a 2D grid. It is closely related to Geary's C,
    but not identical. More info can be found `here <https://en.wikipedia.org/wiki/Moran%27s_I>`_.

    .. math::

        I =
            \frac{
                N \sum_{i, j} w_{i, j} z_{i} z_{j}
            }{
                S_{0} \sum_{i} z_{i}^{2}
            }

    Params
    ------
    adata_or_graph
        AnnData object containing a graph (see ``use_graph``) or the graph itself.
        See the examples for more info.
    vals
        Values to calculate Moran's I for. If this is two dimensional, should
        be of shape `(n_features, n_cells)`. Otherwise should be of shape
        `(n_cells,)`. This matrix can be selected from elements of the anndata
        object by using key word arguments: `layer`, `obsm`, `obsp`, or
        `use_raw`.
    use_graph
        Key to use for graph in anndata object. If not provided, default
        neighbors connectivities will be used instead.
    layer
        Key for `adata.layers` to choose `vals`.
    obsm
        Key for `adata.obsm` to choose `vals`.
    obsp
        Key for `adata.obsp` to choose `vals`.
    use_raw
        Whether to use `adata.raw.X` for `vals`.

    Returns
    -------
    If vals is two dimensional, returns a 1 dimensional ndarray array. Returns
    a scalar if `vals` is 1d.

    Examples
    --------
    Calculate Moran’s I for each components of a dimensionality reduction:

    .. code:: python

        import scanpy as sc, numpy as np

        pbmc = sc.datasets.pbmc68k_processed()
        pc_c = sc.metrics.morans_i(pbmc, obsm="X_pca")

    It's equivalent to call the function directly on the underlying arrays:

    .. code:: python

        alt = sc.metrics.morans_i(pbmc.obsp["connectivities"], pbmc.obsm["X_pca"].T)
        np.testing.assert_array_equal(pc_c, alt)

    """
    adata = cast("AnnData", adata_or_graph)
    g = _get_graph(adata, use_graph=use_graph)
    if vals is None:
        vals = _get_obs_rep(adata, use_raw=use_raw, layer=layer, obsm=obsm, obsp=obsp).T
    return morans_i(g, vals)


@_register_union(morans_i, CSRBase)
def _morans_i(graph: CSRBase, /, vals: _Vals) -> NDArray:
    return _MoransI(graph, vals)()


class _MoransI(_SparseMetric):
    name = "Moran’s I"

    def mtx(self, vals_het: NDArray | CSRBase, /) -> NDArray:
        g_parts = (self.graph.data, self.graph.indices, self.graph.indptr)
        if isinstance(vals_het, np.ndarray):
            return _morans_i_mtx(*g_parts, vals_het)
        v_parts = (vals_het.data, vals_het.indices, vals_het.indptr)
        return _morans_i_mtx_csr(*g_parts, *v_parts, vals_het.shape)

    def vec(self) -> np.float64:
        w = self.graph.data.sum()
        g_parts = (self.graph.data, self.graph.indices, self.graph.indptr)
        return _morans_i_vec_w(*g_parts, self._vals, w)


###############################################################################
# Calculation
###############################################################################
# This is done in a very similar way to gearys_c. See notes there for details.


@numba.njit(cache=True, parallel=False)  # noqa: TID251
def _morans_i_vec_w(
    g_data: np.ndarray,
    g_indices: np.ndarray,
    g_indptr: np.ndarray,
    x: np.ndarray,
    w: np.float64,
) -> np.float64:
    z = x - x.mean()
    z2ss = (z * z).sum()
    n = len(x)
    inum = 0.0

    for i in numba.prange(n):
        s = slice(g_indptr[i], g_indptr[i + 1])
        i_indices = g_indices[s]
        i_data = g_data[s]
        inum += (i_data * z[i_indices]).sum() * z[i]

    return len(x) / w * inum / z2ss


@numba.njit(cache=True, parallel=False)  # noqa: TID251
def _morans_i_vec_w_sparse(  # noqa: PLR0917
    g_data: np.ndarray,
    g_indices: np.ndarray,
    g_indptr: np.ndarray,
    x_data: np.ndarray,
    x_indices: np.ndarray,
    n: int,
    w: np.float64,
) -> np.float64:
    x = np.zeros(n, dtype=x_data.dtype)
    x[x_indices] = x_data
    return _morans_i_vec_w(g_data, g_indices, g_indptr, x, w)


@njit
def _morans_i_mtx(
    g_data: np.ndarray,
    g_indices: np.ndarray,
    g_indptr: np.ndarray,
    x: np.ndarray,
) -> np.ndarray:
    m, n = x.shape
    assert n == len(g_indptr) - 1
    w = g_data.sum()
    out = np.zeros(m, dtype=np.float64)
    for k in numba.prange(m):
        x_vec = x[k, :]
        out[k] = _morans_i_vec_w(g_data, g_indices, g_indptr, x_vec, w)
    return out


@njit
def _morans_i_mtx_csr(  # noqa: PLR0917
    g_data: np.ndarray,
    g_indices: np.ndarray,
    g_indptr: np.ndarray,
    x_data: np.ndarray,
    x_indices: np.ndarray,
    x_indptr: np.ndarray,
    x_shape: tuple,
) -> np.ndarray:
    m, n = x_shape
    w = g_data.sum()
    out = np.zeros(m, dtype=np.float64)
    x_data_list = np.split(x_data, x_indptr[1:-1])
    x_indices_list = np.split(x_indices, x_indptr[1:-1])
    for k in numba.prange(m):
        out[k] = _morans_i_vec_w_sparse(
            g_data,
            g_indices,
            g_indptr,
            x_data_list[k],
            x_indices_list[k],
            n,
            w,
        )
    return out



"""Analysis tools."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ._dendrogram import dendrogram
from ._diffmap import diffmap
from ._dpt import dpt
from ._draw_graph import draw_graph
from ._embedding_density import embedding_density
from ._ingest import (
    Ingest,  # noqa: F401
    ingest,
)
from ._leiden import leiden
from ._louvain import louvain
from ._marker_gene_overlap import marker_gene_overlap
from ._paga import (
    paga,
    paga_compare_paths,  # noqa: F401
    paga_degrees,  # noqa: F401
    paga_expression_entropies,  # noqa: F401
)
from ._rank_genes_groups import filter_rank_genes_groups, rank_genes_groups
from ._score_genes import score_genes, score_genes_cell_cycle
from ._sim import sim
from ._tsne import tsne
from ._umap import umap

if TYPE_CHECKING:
    from typing import Any


def __getattr__(name: str) -> Any:
    if name == "pca":
        from ..preprocessing import pca

        return pca
    raise AttributeError(name)


__all__ = [
    "dendrogram",
    "diffmap",
    "dpt",
    "draw_graph",
    "embedding_density",
    "filter_rank_genes_groups",
    "ingest",
    "leiden",
    "louvain",
    "marker_gene_overlap",
    "paga",
    "rank_genes_groups",
    "score_genes",
    "score_genes_cell_cycle",
    "sim",
    "tsne",
    "umap",
]



from __future__ import annotations

import numpy as np
import pytest

from scanpy.tools import filter_rank_genes_groups, rank_genes_groups
from testing.scanpy._helpers.data import pbmc68k_reduced

NAMES_NO_REF = [
    ["CD3D", "ITM2A", "CD3D", "CCL5", "CD7", "nan", "CD79A", "nan", "NKG7", "LYZ"],
    ["CD3E", "CD3D", "nan", "NKG7", "CD3D", "AIF1", "CD79B", "nan", "GNLY", "CST3"],
    ["IL32", "RPL39", "nan", "CST7", "nan", "nan", "nan", "SNHG7", "CD7", "nan"],
    ["nan", "SRSF7", "IL32", "GZMA", "nan", "LST1", "IGJ", "nan", "CTSW", "nan"],
    ["nan", "nan", "CD2", "CTSW", "CD8B", "TYROBP", "ISG20", "SNHG8", "GZMB", "nan"],
]

NAMES_REF = [
    ["CD3D", "ITM2A", "CD3D", "nan", "CD3D", "nan", "CD79A", "nan", "CD7"],
    ["nan", "nan", "nan", "CD3D", "nan", "AIF1", "nan", "nan", "NKG7"],
    ["nan", "nan", "nan", "NKG7", "nan", "FCGR3A", "ISG20", "SNHG7", "CTSW"],
    ["nan", "CD3D", "nan", "CCL5", "CD7", "nan", "CD79B", "nan", "GNLY"],
    ["CD3E", "IL32", "nan", "IL32", "CD27", "FCER1G", "nan", "nan", "nan"],
]

NAMES_NO_REF_COMPARE_ABS = [
    [
        *("CD3D", "ITM2A", "HLA-DRB1", "CCL5", "HLA-DPA1"),
        *("nan", "CD79A", "nan", "NKG7", "LYZ"),
    ],
    [
        *("HLA-DPA1", "nan", "CD3D", "NKG7", "HLA-DRB1"),
        *("AIF1", "CD79B", "nan", "GNLY", "CST3"),
    ],
    [
        *("nan", "PSAP", "CD74", "CST7", "CD74"),
        *("PSAP", "FCER1G", "SNHG7", "CD7", "HLA-DRA"),
    ],
    [
        *("IL32", "nan", "HLA-DRB5", "GZMA", "HLA-DRB5"),
        *("LST1", "nan", "nan", "CTSW", "HLA-DRB1"),
    ],
    [
        *("nan", "FCER1G", "HLA-DPB1", "CTSW", "HLA-DPB1"),
        *("TYROBP", "TYROBP", "S100A10", "GZMB", "HLA-DPA1"),
    ],
]


EXPECTED = {
    ("Dendritic", False): np.array(NAMES_REF),
    ("rest", False): np.array(NAMES_NO_REF),
    ("rest", True): np.array(NAMES_NO_REF_COMPARE_ABS),
}


@pytest.mark.parametrize(
    ("reference", "pts", "abs"),
    [
        pytest.param("Dendritic", False, False, id="ref-no_pts-no_abs"),
        pytest.param("rest", False, False, id="rest-no_pts-no_abs"),
        pytest.param("rest", True, False, id="rest-pts-no_abs"),
        pytest.param("rest", True, True, id="rest-pts-abs"),
    ],
)
def test_filter_rank_genes_groups(reference, pts, abs):
    adata = pbmc68k_reduced()

    rank_genes_groups(
        adata,
        "bulk_labels",
        reference=reference,
        pts=pts,
        method="wilcoxon",
        rankby_abs=abs,
        n_genes=5,
    )
    if abs:
        filter_rank_genes_groups(
            adata,
            compare_abs=True,
            min_in_group_fraction=-1,
            max_out_group_fraction=1,
            min_fold_change=3.1,
        )
    else:
        filter_rank_genes_groups(
            adata,
            min_in_group_fraction=0.25,
            min_fold_change=1,
            max_out_group_fraction=0.5,
        )

    assert np.array_equal(
        EXPECTED[reference, abs],
        np.array(adata.uns["rank_genes_groups_filtered"]["names"].tolist()),
    )



from __future__ import annotations

import pickle
import string
from contextlib import nullcontext
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pytest
from anndata import AnnData
from anndata.utils import asarray
from scipy import sparse

import scanpy as sc
from scanpy._compat import CSBase
from scanpy._utils.random import random_str
from testing.scanpy._helpers.data import paul15

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Literal

    from scanpy._compat import CSRBase


HERE = Path(__file__).parent
DATA_PATH = HERE / "_data"

_create_random_gene_names = partial(random_str, alphabet=string.ascii_uppercase)
"""Create a bunch of random gene names (just CAPS letters)."""


def _create_sparse_nan_matrix(rows, cols, percent_zero, percent_nan) -> CSRBase:
    """Create a sparse matrix with certain amounts of NaN and Zeros."""
    arr = (
        np.random.randint(0, 1000, rows * cols).reshape((rows, cols)).astype("float32")
    )
    maskzero = np.random.rand(rows, cols) < percent_zero
    masknan = np.random.rand(rows, cols) < percent_nan
    if np.any(maskzero):
        arr[maskzero] = 0
    if np.any(masknan):
        arr[masknan] = np.nan
    return sparse.csr_matrix(arr)  # noqa: TID251


def _create_adata(n_obs: int, n_var: int, p_zero: float, p_nan: float) -> AnnData:
    """Create an AnnData with random data, sparseness and some NaN values."""
    x = _create_sparse_nan_matrix(n_obs, n_var, p_zero, p_nan)
    adata = AnnData(x)
    gene_names = _create_random_gene_names(n_var, length=6)
    adata.var_names = gene_names.reshape(n_var)  # can be unsized
    return adata


def test_score_with_reference():
    """Checks if score_genes output agrees with pre-computed reference values.

    The reference values had been generated using the same code
    and stored as a pickle object in `./data`.
    """
    adata = paul15()
    sc.pp.normalize_per_cell(adata, counts_per_cell_after=10000)
    sc.pp.scale(adata)

    sc.tl.score_genes(adata, gene_list=adata.var_names[:100], score_name="Test")
    with (DATA_PATH / "score_genes_reference_paul2015.pkl").open("rb") as file:
        reference = pickle.load(file)
    # np.testing.assert_allclose(reference, adata.obs["Test"].to_numpy())
    np.testing.assert_array_equal(reference, adata.obs["Test"].to_numpy())


def test_add_score():
    """Check the dtype of the scores and that non-existing genes get ignored."""
    # TODO: write a test that costs less resources and is more meaningful
    adata = _create_adata(100, 1000, p_zero=0, p_nan=0)

    sc.pp.normalize_per_cell(adata, counts_per_cell_after=1e4)
    sc.pp.log1p(adata)

    # the actual genes names are all 6 letters
    # create some non-exstisting names with 7 letters:
    non_existing_genes = _create_random_gene_names(3, length=7)
    some_genes = np.r_[
        np.unique(np.random.choice(adata.var_names, 10)), np.unique(non_existing_genes)
    ]
    sc.tl.score_genes(adata, some_genes, score_name="Test")
    assert adata.obs["Test"].dtype == "float64"


@pytest.mark.parametrize("axis", [0, 1])
@pytest.mark.parametrize(
    "mk_arr",
    [
        pytest.param(
            lambda: _create_sparse_nan_matrix(60, 50, percent_zero=0.3, percent_nan=0),
            id="sparse-no_nan",
        ),
        pytest.param(
            lambda: _create_sparse_nan_matrix(
                60, 50, percent_zero=0.3, percent_nan=0.3
            ),
            id="sparse-some_nan",
        ),
        pytest.param(
            lambda: np.full((10, 1), np.nan),
            marks=pytest.mark.filterwarnings(
                "ignore:Mean of empty slice:RuntimeWarning",
                "ignore:invalid value encountered in divide:RuntimeWarning",
            ),
            id="dense-all_nan",
        ),
    ],
)
def test_sparse_nanmean(
    mk_arr: Callable[[], CSBase | np.ndarray], axis: Literal[0, 1]
) -> None:
    """Check that _sparse_nanmean() is equivalent to np.nanmean()."""
    from scanpy.tools._score_genes import _sparse_nanmean

    arr_or_mat = mk_arr()
    arr = asarray(arr_or_mat)
    mat = sparse.csr_matrix(arr) if not isinstance(arr, CSBase) else arr  # noqa: TID251
    np.testing.assert_allclose(
        np.nanmean(arr, axis), np.array(_sparse_nanmean(mat, axis)).flatten()
    )


def test_sparse_nanmean_on_dense_matrix():
    """TypeError must be thrown when calling _sparse_nanmean with a dense matrix."""
    from scanpy.tools._score_genes import _sparse_nanmean

    with pytest.raises(TypeError):
        _sparse_nanmean(np.random.rand(4, 5), 0)


def test_score_genes_sparse_vs_dense():
    """score_genes() should give the same result for dense and sparse matrices."""
    adata_sparse = _create_adata(100, 1000, p_zero=0.3, p_nan=0.3)

    adata_dense = adata_sparse.copy()
    adata_dense.X = adata_dense.X.toarray()

    gene_set = adata_dense.var_names[:10]

    sc.tl.score_genes(adata_sparse, gene_list=gene_set, score_name="Test")
    sc.tl.score_genes(adata_dense, gene_list=gene_set, score_name="Test")

    np.testing.assert_allclose(
        adata_sparse.obs["Test"].values, adata_dense.obs["Test"].values
    )


def test_score_genes_deplete():
    """Deplete some cells from a set of genes.

    Their score should be <0 since the sum of markers is 0 and
    the sum of random genes is >=0.

    Check that for both sparse and dense matrices.
    """
    adata_sparse = _create_adata(100, 1000, p_zero=0.3, p_nan=0.3)

    adata_dense = adata_sparse.copy()
    adata_dense.X = adata_dense.X.toarray()

    # here's an arbitary gene set
    gene_set = adata_dense.var_names[:10]

    for adata in [adata_sparse, adata_dense]:
        # deplete these genes in 50 cells,
        ix_obs = np.random.choice(adata.shape[0], 50)
        adata[ix_obs][:, gene_set].X = 0

        sc.tl.score_genes(adata, gene_list=gene_set, score_name="Test")
        scores = adata.obs["Test"].values

        np.testing.assert_array_less(scores[ix_obs], 0)


def test_npnanmean_vs_sparsemean(monkeypatch):
    """Another check that _sparsemean behaves like np.nanmean.

    monkeypatch the _score_genes._sparse_nanmean function to np.nanmean
    and check that the result is the same as the non-patched (i.e. sparse_nanmean)
    function
    """
    adata = _create_adata(100, 1000, p_zero=0.3, p_nan=0.3)
    gene_set = adata.var_names[:10]

    # the unpatched, i.e. _sparse_nanmean version
    sc.tl.score_genes(adata, gene_list=gene_set, score_name="Test")
    sparse_scores = adata.obs["Test"].values.tolist()

    # now patch _sparse_nanmean by np.nanmean inside sc.tools
    def mock_fn(x: CSRBase, axis: Literal[0, 1]):
        return np.nanmean(x.toarray(), axis, dtype="float64")

    monkeypatch.setattr(sc.tl._score_genes, "_sparse_nanmean", mock_fn)
    sc.tl.score_genes(adata, gene_list=gene_set, score_name="Test")
    dense_scores = adata.obs["Test"].values

    np.testing.assert_allclose(sparse_scores, dense_scores)


def test_missing_genes():
    adata = _create_adata(100, 1000, p_zero=0, p_nan=0)
    # These genes have a different length of name
    non_extant_genes = _create_random_gene_names(3, length=7)

    with pytest.raises(ValueError, match=r"No valid genes were passed for scoring"):
        sc.tl.score_genes(adata, non_extant_genes)


def test_one_gene():
    # https://github.com/scverse/scanpy/issues/1395
    adata = _create_adata(100, 1000, p_zero=0, p_nan=0)
    sc.tl.score_genes(adata, [adata.var_names[0]])


def test_use_raw_none() -> None:
    adata = _create_adata(100, 1000, p_zero=0, p_nan=0)
    adata_raw = adata.copy()
    adata_raw.var_names = [str(i) for i in range(adata_raw.n_vars)]
    adata.raw = adata_raw

    sc.tl.score_genes(adata, adata_raw.var_names[:3], use_raw=None)


def test_layer():
    adata = _create_adata(100, 1000, p_zero=0, p_nan=0)

    sc.pp.normalize_per_cell(adata, counts_per_cell_after=1e4)
    sc.pp.log1p(adata)

    # score X
    gene_set = adata.var_names[:10]
    sc.tl.score_genes(adata, gene_set, score_name="X_score")
    # score layer (`del` makes sure it actually uses the layer)
    adata.layers["test"] = adata.X.copy()
    adata.raw = adata
    del adata.X
    sc.tl.score_genes(adata, gene_set, score_name="test_score", layer="test")

    np.testing.assert_array_equal(adata.obs["X_score"], adata.obs["test_score"])


@pytest.mark.parametrize("gene_pool", [[], ["foo", "bar"]])
def test_invalid_gene_pool(gene_pool):
    adata = _create_adata(100, 1000, p_zero=0, p_nan=0)

    with pytest.raises(ValueError, match="reference set"):
        sc.tl.score_genes(adata, adata.var_names[:3], gene_pool=gene_pool)


def test_no_control_gene():
    np.random.seed(0)
    adata = _create_adata(100, 1, p_zero=0, p_nan=0)

    with pytest.raises(RuntimeError, match="No control genes found"):
        sc.tl.score_genes(adata, adata.var_names[:1], ctrl_size=1)


@pytest.mark.parametrize(
    "ctrl_as_ref", [True, False], ids=["ctrl_as_ref", "no_ctrl_as_ref"]
)
def test_gene_list_is_control(*, ctrl_as_ref: bool):
    np.random.seed(0)
    adata = sc.datasets.blobs(n_variables=10, n_observations=100, n_centers=20)
    adata.var_names = "g" + adata.var_names
    with (
        pytest.raises(RuntimeError, match=r"No control genes found in any cut")
        if ctrl_as_ref
        else nullcontext()
    ):
        sc.tl.score_genes(
            adata, gene_list="g3", ctrl_size=1, n_bins=5, ctrl_as_ref=ctrl_as_ref
        )



"""Shared docstrings for experimental function parameters."""

from __future__ import annotations

doc_adata = """\
adata
    The annotated data matrix of shape `n_obs` × `n_vars`.
    Rows correspond to cells and columns to genes.
"""

doc_dist_params = """\
theta
    The negative binomial overdispersion parameter `theta` for Pearson residuals.
    Higher values correspond to less overdispersion \
    (`var = mean + mean^2/theta`), and `theta=np.inf` corresponds to a Poisson model.
clip
    Determines if and how residuals are clipped:

    * If `None`, residuals are clipped to the interval \
    `[-sqrt(n_obs), sqrt(n_obs)]`, where `n_obs` is the number of cells in the dataset (default behavior).
    * If any scalar `c`, residuals are clipped to the interval `[-c, c]`. Set \
    `clip=np.inf` for no clipping.
"""

doc_check_values = """\
check_values
    If `True`, checks if counts in selected layer are integers as expected by this
    function, and return a warning if non-integers are found. Otherwise, proceed
    without checking. Setting this to `False` can speed up code for large datasets.
"""

doc_layer = """\
layer
    Layer to use as input instead of `X`. If `None`, `X` is used.
"""

doc_subset = """\
subset
    Inplace subset to highly-variable genes if `True` otherwise merely indicate
    highly variable genes.
"""

doc_genes_batch_chunk = """\
n_top_genes
    Number of highly-variable genes to keep. Mandatory if `flavor='seurat_v3'` or
    `flavor='pearson_residuals'`.
batch_key
    If specified, highly-variable genes are selected within each batch separately
    and merged. This simple process avoids the selection of batch-specific genes
    and acts as a lightweight batch correction method. Genes are first sorted by
    how many batches they are a HVG. If `flavor='pearson_residuals'`, ties are
    broken by the median rank (across batches) based on within-batch residual
    variance.
chunksize
    If `flavor='pearson_residuals'`, this dertermines how many genes are processed at
    once while computing the residual variance. Choosing a smaller value will reduce
    the required memory.
"""

doc_pca_chunk = """\
n_comps
    Number of principal components to compute in the PCA step.
random_state
    Random seed for setting the initial states for the optimization in the PCA step.
kwargs_pca
    Dictionary of further keyword arguments passed on to `scanpy.pp.pca()`.
"""

doc_inplace = """\
inplace
    If `True`, update `adata` with results. Otherwise, return results. See below for
    details of what is returned.
"""

doc_copy = """\
copy
    If `True`, the function runs on a copy of the input object and returns the
    modified copy. Otherwise, the input object is modified direcly. Not compatible
    with `inplace=False`.
"""



from __future__ import annotations

import warnings
from typing import TypeVar

import numpy as np
from numpy.typing import NDArray
from scipy import sparse

from .._compat import CSRBase
from ._common import (
    _get_indices_distances_from_dense_matrix,
    _get_indices_distances_from_sparse_matrix,
)

D = TypeVar("D", NDArray[np.float32], CSRBase)


def gauss(distances: D, n_neighbors: int, *, knn: bool) -> D:  # noqa: PLR0912
    """Derive gaussian connectivities between data points from their distances.

    Parameters
    ----------
    distances
        The input matrix of distances between data points.
    n_neighbors
        The number of nearest neighbors to consider.
    knn
        Specify if the distances have been restricted to k nearest neighbors.

    """
    # init distances
    if isinstance(distances, CSRBase):
        d_sq = distances.power(2)
        indices, distances_sq = _get_indices_distances_from_sparse_matrix(
            d_sq, n_neighbors
        )
    else:
        assert isinstance(distances, np.ndarray)
        d_sq = np.power(distances, 2)
        indices, distances_sq = _get_indices_distances_from_dense_matrix(
            d_sq, n_neighbors
        )

    # exclude the first point, the 0th neighbor
    indices = indices[:, 1:]
    distances_sq = distances_sq[:, 1:]

    # choose sigma, the heuristic here doesn't seem to make much of a difference,
    # but is used to reproduce the figures of Haghverdi et al. (2016)
    if isinstance(distances, CSRBase):
        # as the distances are not sorted
        # we have decay within the n_neighbors first neighbors
        sigmas_sq = np.median(distances_sq, axis=1)
    else:
        # the last item is already in its sorted position through argpartition
        # we have decay beyond the n_neighbors neighbors
        sigmas_sq = distances_sq[:, -1] / 4
    sigmas = np.sqrt(sigmas_sq)

    # compute the symmetric weight matrix
    if not isinstance(distances, CSRBase):
        num = 2 * np.multiply.outer(sigmas, sigmas)
        den = np.add.outer(sigmas_sq, sigmas_sq)
        w = np.sqrt(num / den) * np.exp(-d_sq / den)
        # make the weight matrix sparse
        if not knn:
            mask = w > 1e-14
            w[~mask] = 0
        else:
            # restrict number of neighbors to ~k
            # build a symmetric mask
            mask = np.zeros(d_sq.shape, dtype=bool)
            for i, row in enumerate(indices):
                mask[i, row] = True
                for j in row:
                    if i not in set(indices[j]):
                        w[j, i] = w[i, j]
                        mask[j, i] = True
            # set all entries that are not nearest neighbors to zero
            w[~mask] = 0
    else:
        assert isinstance(d_sq, CSRBase)
        # need to copy the distance matrix here; what follows is inplace
        w = d_sq.copy()
        for i in range(len(d_sq.indptr[:-1])):
            row = d_sq.indices[d_sq.indptr[i] : d_sq.indptr[i + 1]]
            num = 2 * sigmas[i] * sigmas[row]
            den = sigmas_sq[i] + sigmas_sq[row]
            w.data[d_sq.indptr[i] : d_sq.indptr[i + 1]] = np.sqrt(num / den) * np.exp(
                -d_sq.data[d_sq.indptr[i] : d_sq.indptr[i + 1]] / den
            )
        w = w.tolil()
        for i, row in enumerate(indices):
            for j in row:
                if i not in set(indices[j]):
                    w[j, i] = w[i, j]
        w = w.tocsr()

    return w


def umap(
    knn_indices: NDArray[np.int32 | np.int64],
    knn_dists: NDArray[np.float32 | np.float64],
    *,
    n_obs: int,
    n_neighbors: int,
    set_op_mix_ratio: float = 1.0,
    local_connectivity: float = 1.0,
) -> CSRBase:
    """Wrap for `umap.fuzzy_simplicial_set` :cite:p:`McInnes2018`.

    Given a set of data X, a neighborhood size, and a measure of distance
    compute the fuzzy simplicial set (here represented as a fuzzy graph in
    the form of a sparse matrix) associated to the data. This is done by
    locally approximating geodesic distance at each point, creating a fuzzy
    simplicial set for each such point, and then combining all the local
    fuzzy simplicial sets into a global one via a fuzzy union.
    """
    with warnings.catch_warnings():
        # umap 0.5.0
        warnings.filterwarnings("ignore", message=r"Tensorflow not installed")
        from umap.umap_ import fuzzy_simplicial_set

    x = sparse.coo_matrix((n_obs, 1))
    connectivities, _sigmas, _rhos = fuzzy_simplicial_set(
        x,
        n_neighbors,
        None,
        None,
        knn_indices=knn_indices,
        knn_dists=knn_dists,
        set_op_mix_ratio=set_op_mix_ratio,
        local_connectivity=local_connectivity,
    )

    return connectivities.tocsr()



"""External preprocessing functions."""

from __future__ import annotations

from sklearn.utils import deprecated

from ...preprocessing import _scrublet
from ._bbknn import bbknn
from ._dca import dca
from ._harmony_integrate import harmony_integrate
from ._hashsolo import hashsolo
from ._magic import magic
from ._mnn_correct import mnn_correct
from ._scanorama_integrate import scanorama_integrate

scrublet = deprecated("Import from sc.pp instead")(_scrublet.scrublet)
scrublet_simulate_doublets = deprecated("Import from sc.pp instead")(
    _scrublet.scrublet_simulate_doublets
)

__all__ = [
    "bbknn",
    "dca",
    "harmony_integrate",
    "hashsolo",
    "magic",
    "mnn_correct",
    "scanorama_integrate",
]



"""Computes a dendrogram based on a given categorical observation."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd
from pandas.api.types import CategoricalDtype

from .. import logging as logg
from .._compat import old_positionals
from .._utils import _doc_params, raise_not_implemented_error_if_backed_type
from ..neighbors._doc import doc_n_pcs, doc_use_rep
from ._utils import _choose_representation

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any

    from anndata import AnnData


@old_positionals(
    "n_pcs",
    "use_rep",
    "var_names",
    "use_raw",
    "cor_method",
    "linkage_method",
    "optimal_ordering",
    "key_added",
    "inplace",
)
@_doc_params(n_pcs=doc_n_pcs, use_rep=doc_use_rep)
def dendrogram(  # noqa: PLR0913
    adata: AnnData,
    groupby: str | Sequence[str],
    *,
    n_pcs: int | None = None,
    use_rep: str | None = None,
    var_names: Sequence[str] | None = None,
    use_raw: bool | None = None,
    cor_method: str = "pearson",
    linkage_method: str = "complete",
    optimal_ordering: bool = False,
    key_added: str | None = None,
    inplace: bool = True,
) -> dict[str, Any] | None:
    """Compute a hierarchical clustering for the given `groupby` categories.

    By default, the PCA representation is used unless `.X`
    has less than 50 variables.

    Alternatively, a list of `var_names` (e.g. genes) can be given.

    Average values of either `var_names` or components are used
    to compute a correlation matrix.

    The hierarchical clustering can be visualized using
    :func:`scanpy.pl.dendrogram` or multiple other visualizations
    that can include a dendrogram: :func:`~scanpy.pl.matrixplot`,
    :func:`~scanpy.pl.heatmap`, :func:`~scanpy.pl.dotplot`,
    and :func:`~scanpy.pl.stacked_violin`.

    .. note::
        The computation of the hierarchical clustering is based on predefined
        groups and not per cell. The correlation matrix is computed using by
        default pearson but other methods are available.

    Parameters
    ----------
    adata
        Annotated data matrix
    {n_pcs}
    {use_rep}
    var_names
        List of var_names to use for computing the hierarchical clustering.
        If `var_names` is given, then `use_rep` and `n_pcs` are ignored.
    use_raw
        Only when `var_names` is not None.
        Use `raw` attribute of `adata` if present.
    cor_method
        Correlation method to use.
        Options are 'pearson', 'kendall', and 'spearman'
    linkage_method
        Linkage method to use. See :func:`scipy.cluster.hierarchy.linkage`
        for more information.
    optimal_ordering
        Same as the optimal_ordering argument of :func:`scipy.cluster.hierarchy.linkage`
        which reorders the linkage matrix so that the distance between successive
        leaves is minimal.
    key_added
        By default, the dendrogram information is added to
        `.uns[f'dendrogram_{{groupby}}']`.
        Notice that the `groupby` information is added to the dendrogram.
    inplace
        If `True`, adds dendrogram information to `adata.uns[key_added]`,
        else this function returns the information.

    Returns
    -------
    Returns `None` if `inplace=True`, else returns a `dict` with dendrogram information. Sets the following field if `inplace=True`:

    `adata.uns[f'dendrogram_{{group_by}}' | key_added]` : :class:`dict`
        Dendrogram information.

    Examples
    --------
    >>> import scanpy as sc
    >>> adata = sc.datasets.pbmc68k_reduced()
    >>> sc.tl.dendrogram(adata, groupby="bulk_labels")
    >>> sc.pl.dendrogram(adata, groupby="bulk_labels")  # doctest: +SKIP
    <Axes: >
    >>> markers = ["C1QA", "PSAP", "CD79A", "CD79B", "CST3", "LYZ"]
    >>> sc.pl.dotplot(adata, markers, groupby="bulk_labels", dendrogram=True)

    """
    raise_not_implemented_error_if_backed_type(adata.X, "dendrogram")
    if isinstance(groupby, str):
        # if not a list, turn into a list
        groupby = [groupby]
    for group in groupby:
        if group not in adata.obs:
            msg = (
                "groupby has to be a valid observation. "
                f"Given value: {group}, valid observations: {adata.obs.columns.tolist()}"
            )
            raise ValueError(msg)
        if not isinstance(adata.obs[group].dtype, CategoricalDtype):
            msg = (
                "groupby has to be a categorical observation. "
                f"Given value: {group}, Column type: {adata.obs[group].dtype}"
            )
            raise ValueError(msg)

    if var_names is None:
        rep_df = pd.DataFrame(
            _choose_representation(adata, use_rep=use_rep, n_pcs=n_pcs)
        )
        categorical = adata.obs[groupby[0]]
        if len(groupby) > 1:
            for group in groupby[1:]:
                # create new category by merging the given groupby categories
                categorical = (
                    categorical.astype(str) + "_" + adata.obs[group].astype(str)
                ).astype("category")
        categorical.name = "_".join(groupby)

        rep_df.set_index(categorical, inplace=True)
        categories: pd.Index = rep_df.index.categories
    else:
        gene_names = adata.raw.var_names if use_raw else adata.var_names
        from ..plotting._anndata import _prepare_dataframe

        categories, rep_df = _prepare_dataframe(
            adata, gene_names, groupby, use_raw=use_raw
        )

    # aggregate values within categories using 'mean'
    mean_df = (
        rep_df.groupby(level=0, observed=True)
        .mean()
        .loc[categories]  # Fixed ordering for pandas < 2
    )

    import scipy.cluster.hierarchy as sch
    from scipy.spatial import distance

    corr_matrix = mean_df.T.corr(method=cor_method).clip(-1, 1)
    corr_condensed = distance.squareform(1 - corr_matrix)
    z_var = sch.linkage(
        corr_condensed, method=linkage_method, optimal_ordering=optimal_ordering
    )
    dendro_info = sch.dendrogram(z_var, labels=list(categories), no_plot=True)

    dat = dict(
        linkage=z_var,
        groupby=groupby,
        use_rep=use_rep,
        cor_method=cor_method,
        linkage_method=linkage_method,
        categories_ordered=dendro_info["ivl"],
        categories_idx_ordered=dendro_info["leaves"],
        dendrogram_info=dendro_info,
        correlation_matrix=corr_matrix.values,
    )

    if inplace:
        if key_added is None:
            key_added = f"dendrogram_{'_'.join(groupby)}"
        logg.info(f"Storing dendrogram info using `.uns[{key_added!r}]`")
        adata.uns[key_added] = dat
    else:
        return dat



from __future__ import annotations

import pickle
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pytest
from anndata import AnnData
from numpy.random import binomial, negative_binomial, seed
from packaging.version import Version
from scipy.stats import mannwhitneyu

import scanpy as sc
from scanpy._compat import pkg_version
from scanpy._utils import elem_mul, select_groups
from scanpy.get import rank_genes_groups_df
from scanpy.tools import rank_genes_groups
from scanpy.tools._rank_genes_groups import _RankGenes
from testing.scanpy._helpers import random_mask
from testing.scanpy._helpers.data import pbmc68k_reduced
from testing.scanpy._pytest.params import ARRAY_TYPES, ARRAY_TYPES_MEM

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any

    from numpy.typing import NDArray

HERE = Path(__file__).parent
DATA_PATH = HERE / "_data"


# We test results for a simple generic example
# Tests are conducted for sparse and non-sparse AnnData objects.
# Due to minor changes in multiplication implementation for sparse and non-sparse objects,
# results differ (very) slightly


@pytest.mark.parametrize("array_type", ARRAY_TYPES)
def get_example_data(array_type: Callable[[np.ndarray], Any]) -> AnnData:
    # create test object
    adata = AnnData(
        np.multiply(binomial(1, 0.15, (100, 20)), negative_binomial(2, 0.25, (100, 20)))
    )
    # adapt marker_genes for cluster (so as to have some form of reasonable input
    adata.X[0:10, 0:5] = np.multiply(
        binomial(1, 0.9, (10, 5)), negative_binomial(1, 0.5, (10, 5))
    )

    adata.X = array_type(adata.X)

    # Create cluster according to groups
    adata.obs["true_groups"] = pd.Categorical(
        np.concatenate((np.zeros((10,), dtype=int), np.ones((90,), dtype=int)))
    )

    return adata


def get_true_scores() -> tuple[
    NDArray[np.object_],
    NDArray[np.object_],
    NDArray[np.floating],
    NDArray[np.floating],
]:
    with (DATA_PATH / "objs_t_test.pkl").open("rb") as f:
        true_scores_t_test, true_names_t_test = pickle.load(f)
    with (DATA_PATH / "objs_wilcoxon.pkl").open("rb") as f:
        true_scores_wilcoxon, true_names_wilcoxon = pickle.load(f)

    return (
        true_names_t_test,
        true_names_wilcoxon,
        true_scores_t_test,
        true_scores_wilcoxon,
    )


# TODO: Make dask compatible
@pytest.mark.parametrize("array_type", ARRAY_TYPES_MEM)
def test_results(array_type):
    seed(1234)

    adata = get_example_data(array_type)
    assert adata.raw is None  # Assumption for later checks

    (
        true_names_t_test,
        true_names_wilcoxon,
        true_scores_t_test,
        true_scores_wilcoxon,
    ) = get_true_scores()

    rank_genes_groups(adata, "true_groups", n_genes=20, method="t-test")

    adata.uns["rank_genes_groups"]["names"] = adata.uns["rank_genes_groups"][
        "names"
    ].astype(true_names_t_test.dtype)

    for name in true_scores_t_test.dtype.names:
        assert np.allclose(
            true_scores_t_test[name], adata.uns["rank_genes_groups"]["scores"][name]
        )
    assert np.array_equal(true_names_t_test, adata.uns["rank_genes_groups"]["names"])
    assert adata.uns["rank_genes_groups"]["params"]["use_raw"] is False

    rank_genes_groups(adata, "true_groups", n_genes=20, method="wilcoxon")

    adata.uns["rank_genes_groups"]["names"] = adata.uns["rank_genes_groups"][
        "names"
    ].astype(true_names_wilcoxon.dtype)

    for name in true_scores_t_test.dtype.names:
        assert np.allclose(
            true_scores_wilcoxon[name][:7],
            adata.uns["rank_genes_groups"]["scores"][name][:7],
        )
    assert np.array_equal(
        true_names_wilcoxon[:7], adata.uns["rank_genes_groups"]["names"][:7]
    )
    assert adata.uns["rank_genes_groups"]["params"]["use_raw"] is False


@pytest.mark.parametrize("array_type", ARRAY_TYPES_MEM)
def test_results_layers(array_type):
    seed(1234)

    adata = get_example_data(array_type)
    adata.layers["to_test"] = adata.X.copy()
    adata.X = elem_mul(adata.X, np.random.randint(0, 2, adata.shape, dtype=bool))

    _, _, true_scores_t_test, true_scores_wilcoxon = get_true_scores()

    # Wilcoxon
    rank_genes_groups(
        adata,
        "true_groups",
        method="wilcoxon",
        layer="to_test",
        n_genes=20,
    )
    assert adata.uns["rank_genes_groups"]["params"]["use_raw"] is False
    for name in true_scores_t_test.dtype.names:
        assert np.allclose(
            true_scores_wilcoxon[name][:7],
            adata.uns["rank_genes_groups"]["scores"][name][:7],
        )

    rank_genes_groups(adata, "true_groups", method="wilcoxon", n_genes=20)
    for name in true_scores_t_test.dtype.names:
        assert not np.allclose(
            true_scores_wilcoxon[name][:7],
            adata.uns["rank_genes_groups"]["scores"][name][:7],
        )

    # t-test
    rank_genes_groups(
        adata,
        "true_groups",
        method="t-test",
        layer="to_test",
        use_raw=False,
        n_genes=20,
    )
    for name in true_scores_t_test.dtype.names:
        assert np.allclose(
            true_scores_t_test[name][:7],
            adata.uns["rank_genes_groups"]["scores"][name][:7],
        )

    rank_genes_groups(adata, "true_groups", method="t-test", n_genes=20)
    for name in true_scores_t_test.dtype.names:
        assert not np.allclose(
            true_scores_t_test[name][:7],
            adata.uns["rank_genes_groups"]["scores"][name][:7],
        )


def test_rank_genes_groups_use_raw():
    # https://github.com/scverse/scanpy/issues/1929
    pbmc = pbmc68k_reduced()
    assert pbmc.raw is not None

    sc.tl.rank_genes_groups(pbmc, groupby="bulk_labels", use_raw=True)

    pbmc = pbmc68k_reduced()
    del pbmc.raw
    assert pbmc.raw is None

    with pytest.raises(
        ValueError, match=r"Received `use_raw=True`, but `adata\.raw` is empty"
    ):
        sc.tl.rank_genes_groups(pbmc, groupby="bulk_labels", use_raw=True)


def test_singlets():
    pbmc = pbmc68k_reduced()
    pbmc.obs["louvain"] = pbmc.obs["louvain"].cat.add_categories(["11"])
    pbmc.obs["louvain"][0] = "11"

    with pytest.raises(ValueError, match=rf"Could not calculate statistics.*{'11'}"):
        rank_genes_groups(pbmc, groupby="louvain")


def test_emptycat():
    pbmc = pbmc68k_reduced()
    pbmc.obs["louvain"] = pbmc.obs["louvain"].cat.add_categories(["11"])

    with pytest.raises(ValueError, match=rf"Could not calculate statistics.*{'11'}"):
        rank_genes_groups(pbmc, groupby="louvain")


def test_log1p_save_restore(tmp_path):
    """Tests the sequence log1p→save→load→rank_genes_groups."""
    from anndata import read_h5ad

    pbmc = pbmc68k_reduced()
    sc.pp.log1p(pbmc)

    path = tmp_path / "test.h5ad"
    pbmc.write(path)

    pbmc = read_h5ad(path)

    sc.tl.rank_genes_groups(pbmc, groupby="bulk_labels", use_raw=True)


def test_wilcoxon_symmetry():
    pbmc = pbmc68k_reduced()

    rank_genes_groups(
        pbmc,
        groupby="bulk_labels",
        groups=["CD14+ Monocyte", "Dendritic"],
        reference="Dendritic",
        method="wilcoxon",
        rankby_abs=True,
    )
    assert pbmc.uns["rank_genes_groups"]["params"]["use_raw"] is True

    stats_mono = (
        rank_genes_groups_df(pbmc, group="CD14+ Monocyte")
        .drop(columns="names")
        .to_numpy()
    )

    rank_genes_groups(
        pbmc,
        groupby="bulk_labels",
        groups=["CD14+ Monocyte", "Dendritic"],
        reference="CD14+ Monocyte",
        method="wilcoxon",
        rankby_abs=True,
    )

    stats_dend = (
        rank_genes_groups_df(pbmc, group="Dendritic").drop(columns="names").to_numpy()
    )

    assert np.allclose(np.abs(stats_mono), np.abs(stats_dend))


@pytest.mark.parametrize("reference", [True, False])
def test_wilcoxon_tie_correction(reference):
    pbmc = pbmc68k_reduced()

    groups = ["CD14+ Monocyte", "Dendritic"]
    groupby = "bulk_labels"

    _, groups_masks = select_groups(pbmc, groups, groupby)

    x = pbmc.raw.X[groups_masks[0]].toarray()

    mask_rest = groups_masks[1] if reference else ~groups_masks[0]
    y = pbmc.raw.X[mask_rest].toarray()

    # Handle scipy versions
    if pkg_version("scipy") >= Version("1.7.0"):
        pvals = mannwhitneyu(x, y, use_continuity=False, alternative="two-sided").pvalue
        pvals[np.isnan(pvals)] = 1.0
    else:
        # Backwards compat, to drop once we drop scipy < 1.7
        n_genes = x.shape[1]
        pvals = np.zeros(n_genes)

        for i in range(n_genes):
            try:
                _, pvals[i] = mannwhitneyu(
                    x[:, i], y[:, i], use_continuity=False, alternative="two-sided"
                )
            except ValueError:  # noqa: PERF203
                pvals[i] = 1

    if reference:
        ref = groups[1]
    else:
        ref = "rest"
        groups = groups[:1]

    test_obj = _RankGenes(pbmc, groups, groupby, reference=ref)
    test_obj.compute_statistics("wilcoxon", tie_correct=True)

    np.testing.assert_allclose(test_obj.stats[groups[0]]["pvals"], pvals)


def test_wilcoxon_huge_data(monkeypatch):
    max_size = 300
    adata = pbmc68k_reduced()
    monkeypatch.setattr(sc.tl._rank_genes_groups, "_CONST_MAX_SIZE", max_size)
    rank_genes_groups(adata, groupby="bulk_labels", method="wilcoxon")


@pytest.mark.parametrize(
    ("n_genes_add", "n_genes_out_add"),
    [pytest.param(0, 0, id="equal"), pytest.param(2, 1, id="more")],
)
def test_mask_n_genes(n_genes_add, n_genes_out_add):
    """Check if no. genes in output is correct.

    1. =n_genes when n_genes<sum(mask)
    2. =sum(mask) when n_genes>sum(mask)
    """
    pbmc = pbmc68k_reduced()
    mask_var = np.zeros(pbmc.shape[1]).astype(bool)
    mask_var[:6].fill(True)  # noqa: FBT003
    no_genes = sum(mask_var) - 1

    rank_genes_groups(
        pbmc,
        mask_var=mask_var,
        groupby="bulk_labels",
        groups=["CD14+ Monocyte", "Dendritic"],
        reference="CD14+ Monocyte",
        n_genes=no_genes + n_genes_add,
        method="wilcoxon",
    )

    assert len(pbmc.uns["rank_genes_groups"]["scores"]) == no_genes + n_genes_out_add


def test_mask_not_equal():
    """Check that mask is applied successfully to data set where test statistics are already available (test stats overwritten)."""
    pbmc = pbmc68k_reduced()
    mask_var = random_mask(pbmc.shape[1])
    n_genes = sum(mask_var)

    run = partial(
        rank_genes_groups,
        pbmc,
        groupby="bulk_labels",
        groups=["CD14+ Monocyte", "Dendritic"],
        reference="CD14+ Monocyte",
        method="wilcoxon",
    )

    run(n_genes=n_genes)
    no_mask = pbmc.uns["rank_genes_groups"]["names"]

    run(mask_var=mask_var)
    with_mask = pbmc.uns["rank_genes_groups"]["names"]

    assert not np.array_equal(no_mask, with_mask)



from __future__ import annotations

import inspect
import sys
from functools import wraps
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Literal, ParamSpec, TypeVar, get_args

from .. import logging
from .._compat import old_positionals
from .._singleton import SingletonMeta, documenting
from ..logging import _RootLogger, _set_log_file, _set_log_level
from .verbosity import Verbosity

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from types import UnionType
    from typing import ClassVar, Concatenate, Self, TextIO

    from .verbosity import _VerbosityName

    # Collected from the print_* functions in matplotlib.backends
    _Format = (
        Literal["png", "jpg", "tif", "tiff"]  # noqa: PYI030
        | Literal["pdf", "ps", "eps", "svg", "svgz", "pgf"]
        | Literal["raw", "rgba"]
    )


S = TypeVar("S")
T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")


AnnDataFileFormat = Literal["h5ad", "zarr"]


def _type_check(var: object, name: str, types: type | UnionType):
    if isinstance(var, types):
        return
    if isinstance(types, type):
        possible_types_str = types.__name__
    else:
        type_names = [t.__name__ for t in get_args(types)]
        possible_types_str = f"{', '.join(type_names[:-1])} or {type_names[-1]}"
    msg = f"{name} must be of type {possible_types_str}"
    raise TypeError(msg)


def _type_check_arg2(
    types: type | UnionType,
) -> Callable[[Callable[Concatenate[S, T, P], R]], Callable[Concatenate[S, T, P], R]]:
    def decorator(
        func: Callable[Concatenate[S, T, P], R],
    ) -> Callable[Concatenate[S, T, P], R]:
        @wraps(func)
        def wrapped(self: S, var: T, *args: P.args, **kwargs: P.kwargs) -> R:
            __tracebackhide__ = True
            _type_check(var, func.__name__, types)
            return func(self, var, *args, **kwargs)

        return wrapped

    return decorator


# `type` is only here because of https://github.com/astral-sh/ruff/issues/20225
class SettingsMeta(SingletonMeta, type):
    # logging
    _root_logger: _RootLogger
    _logfile: TextIO
    _verbosity: Verbosity
    # rest
    _n_pcs: int
    _plot_suffix: str
    _file_format_data: AnnDataFileFormat
    _file_format_figs: str
    _autosave: bool
    _autoshow: bool
    _writedir: Path
    _cachedir: Path
    _datasetdir: Path
    _figdir: Path
    _cache_compression: Literal["lzf", "gzip"] | None
    _max_memory: float
    _n_jobs: int
    _categories_to_ignore: list[str]
    _frameon: bool
    """See set_figure_params."""
    _vector_friendly: bool
    """Set to true if you want to include pngs in svgs and pdfs."""
    _low_resolution_warning: bool
    """Print warning when saving a figure with low resolution."""
    _start: float
    """Time when the settings module is first imported."""
    _previous_time: float
    """Variable for timing program parts."""
    _previous_memory_usage: int
    """Stores the previous memory usage."""

    @property
    def verbosity(cls) -> Verbosity:
        """Verbosity level (default :attr:`Verbosity.warning`)."""
        return cls._verbosity

    @verbosity.setter
    def verbosity(cls, verbosity: Verbosity | _VerbosityName | int) -> None:
        try:
            cls._verbosity = (
                Verbosity[verbosity.lower()]
                if isinstance(verbosity, str)
                else Verbosity(verbosity)
            )
        except KeyError:
            msg = (
                f"Cannot set verbosity to {verbosity}. "
                f"Accepted string values are: {Verbosity.__members__.keys()}"
            )
            raise ValueError(msg) from None
        _set_log_level(cls, cls._verbosity.level)

    @property
    def N_PCS(cls) -> int:  # noqa: N802
        """Default number of principal components to use."""
        return cls._n_pcs

    @N_PCS.setter
    @_type_check_arg2(int)
    def N_PCS(cls, n_pcs: int) -> None:  # noqa: N802
        cls._n_pcs = n_pcs

    @property
    def plot_suffix(cls) -> str:
        """Global suffix that is appended to figure filenames."""
        return cls._plot_suffix

    @plot_suffix.setter
    @_type_check_arg2(str)
    def plot_suffix(cls, plot_suffix: str) -> None:
        cls._plot_suffix = plot_suffix

    @property
    def file_format_data(cls) -> AnnDataFileFormat:
        """File format for saving AnnData objects."""
        return cls._file_format_data

    @file_format_data.setter
    @_type_check_arg2(str)
    def file_format_data(cls, file_format: AnnDataFileFormat) -> None:
        if file_format not in (file_format_options := get_args(AnnDataFileFormat)):
            msg = (
                f"Cannot set file_format_data to {file_format}. "
                f"Must be one of {file_format_options}"
            )
            raise ValueError(msg)
        cls._file_format_data: AnnDataFileFormat = file_format

    @property
    def file_format_figs(cls) -> str:
        """File format for saving figures.

        For example `'png'`, `'pdf'` or `'svg'`. Many other formats work as well (see
        :func:`matplotlib.pyplot.savefig`).
        """
        return cls._file_format_figs

    @file_format_figs.setter
    @_type_check_arg2(str)
    def file_format_figs(cls, figure_format: str) -> None:
        cls._file_format_figs = figure_format

    @property
    def autosave(cls) -> bool:
        """Automatically save figures in :attr:`~scanpy.settings.figdir` (default `False`).

        Do not show plots/figures interactively.
        """
        return cls._autosave

    @autosave.setter
    @_type_check_arg2(bool)
    def autosave(cls, autosave: bool) -> None:
        cls._autosave = autosave

    @property
    def autoshow(cls) -> bool:
        """Automatically show figures if `autosave == False` (default `True`).

        There is no need to call the matplotlib pl.show() in this case.
        """
        return cls._autoshow

    @autoshow.setter
    @_type_check_arg2(bool)
    def autoshow(cls, autoshow: bool) -> None:
        cls._autoshow = autoshow

    @property
    def writedir(cls) -> Path:
        """Directory where the function scanpy.write writes to by default."""
        return cls._writedir

    @writedir.setter
    @_type_check_arg2(Path | str)
    def writedir(cls, writedir: Path | str) -> None:
        cls._writedir = Path(writedir)

    @property
    def cachedir(cls) -> Path:
        """Directory for cache files (default `'./cache/'`)."""
        return cls._cachedir

    @cachedir.setter
    @_type_check_arg2(Path | str)
    def cachedir(cls, cachedir: Path | str) -> None:
        cls._cachedir = Path(cachedir)

    @property
    def datasetdir(cls) -> Path:
        """Directory for example :mod:`~scanpy.datasets` (default `'./data/'`)."""
        return cls._datasetdir

    @datasetdir.setter
    @_type_check_arg2(Path | str)
    def datasetdir(cls, datasetdir: Path | str) -> None:
        cls._datasetdir = Path(datasetdir).resolve()

    @property
    def figdir(cls) -> Path:
        """Directory for saving figures (default `'./figures/'`)."""
        return cls._figdir

    @figdir.setter
    @_type_check_arg2(Path | str)
    def figdir(cls, figdir: Path | str) -> None:
        cls._figdir = Path(figdir)

    @property
    def cache_compression(cls) -> Literal["lzf", "gzip"] | None:
        """Compression for `sc.read(..., cache=True)` (default `'lzf'`)."""
        return cls._cache_compression

    @cache_compression.setter
    def cache_compression(
        cls, cache_compression: Literal["lzf", "gzip"] | None
    ) -> None:
        if cache_compression not in {"lzf", "gzip", None}:
            msg = (
                f"`cache_compression` ({cache_compression}) "
                "must be in {'lzf', 'gzip', None}"
            )
            raise ValueError(msg)
        cls._cache_compression = cache_compression

    @property
    def max_memory(cls) -> int | float:
        """Maximum memory usage in Gigabyte.

        Is currently not well respected…
        """
        return cls._max_memory

    @max_memory.setter
    @_type_check_arg2(int | float)
    def max_memory(cls, max_memory: float) -> None:
        cls._max_memory = max_memory

    @property
    def n_jobs(cls) -> int:
        """Default number of jobs/ CPUs to use for parallel computing.

        Set to `-1` in order to use all available cores.
        Not all algorithms support special behavior for numbers < `-1`,
        so make sure to leave this setting as >= `-1`.
        """
        return cls._n_jobs

    @n_jobs.setter
    @_type_check_arg2(int)
    def n_jobs(cls, n_jobs: int) -> None:
        cls._n_jobs = n_jobs

    @property
    def logpath(cls) -> Path | None:
        """The file path `logfile` was set to."""
        return cls._logpath

    @logpath.setter
    @_type_check_arg2(Path | str)
    def logpath(cls, logpath: Path | str | None) -> None:
        if logpath is None:
            cls.logfile = None
            cls._logpath = None
            return
        # set via “file object” branch of logfile.setter
        cls.logfile = Path(logpath).open("a")  # noqa: SIM115
        cls._logpath = Path(logpath)

    @property
    def logfile(cls) -> TextIO:
        """The open file to write logs to.

        Set it to a :class:`~pathlib.Path` or :class:`str` to open a new one.
        The default `None` corresponds to :obj:`sys.stdout` in jupyter notebooks
        and to :obj:`sys.stderr` otherwise.

        For backwards compatibility, setting it to `''` behaves like setting it to `None`.
        """
        return cls._logfile

    @logfile.setter
    def logfile(cls, logfile: Path | str | TextIO | None) -> None:
        if not logfile:  # "" or None
            logfile = cls._default_logfile()
        if isinstance(logfile, Path | str):
            cls.logpath = logfile
            return
        cls._logfile = logfile
        cls._logpath = None
        _set_log_file(cls)

    @property
    def categories_to_ignore(cls) -> list[str]:
        """Categories that are omitted in plotting etc."""
        return cls._categories_to_ignore

    @categories_to_ignore.setter
    def categories_to_ignore(cls, categories_to_ignore: Iterable[str]) -> None:
        categories_to_ignore = list(categories_to_ignore)
        for i, cat in enumerate(categories_to_ignore):
            _type_check(cat, f"categories_to_ignore[{i}]", str)
        cls._categories_to_ignore = categories_to_ignore

    # --------------------------------------------------------------------------------
    # Functions
    # --------------------------------------------------------------------------------

    def set_figure_params(cls, *args, **kwargs) -> None:
        cls._set_figure_params(*args, **kwargs)

    @old_positionals(
        "scanpy",
        "dpi",
        "dpi_save",
        "frameon",
        "vector_friendly",
        "fontsize",
        "figsize",
        "color_map",
        "format",
        "facecolor",
        "transparent",
        "ipython_format",
    )
    def _set_figure_params(  # noqa: PLR0913
        cls,
        *,
        scanpy: bool = True,
        dpi: int = 80,
        dpi_save: int = 150,
        frameon: bool = True,
        vector_friendly: bool = True,
        fontsize: int = 14,
        figsize: int | None = None,
        color_map: str | None = None,
        format: _Format = "pdf",
        facecolor: str | None = None,
        transparent: bool = False,
        ipython_format: str | Iterable[str] = "retina",
    ) -> None:
        """Set resolution/size, styling and format of figures.

        Parameters
        ----------
        scanpy
            Init default values for :obj:`matplotlib.rcParams` suited for Scanpy.
        dpi
            Resolution of rendered figures – this influences the size of figures in notebooks.
        dpi_save
            Resolution of saved figures.
            This should typically be higher to achieve publication quality.
        frameon
            Add frames and axes labels to scatter plots.
        vector_friendly
            Plot scatter plots using `png` backend even when exporting as `pdf` or `svg`.
        fontsize
            Set the fontsize for several `rcParams` entries. Ignored if `scanpy=False`.
        figsize
            Set `rcParams['figure.figsize']`.
        color_map
            Convenience method for setting the default color map. Ignored if `scanpy=False`.
        format
            This sets the default format for saving figures: `file_format_figs`.
        facecolor
            Sets backgrounds via `rcParams['figure.facecolor'] = facecolor` and
            `rcParams['axes.facecolor'] = facecolor`.
        transparent
            Save figures with transparent background.
            Sets `rcParams['savefig.transparent']`.
        ipython_format
            Only concerns the notebook/IPython environment; see
            `matplotlib_inline.backend_inline.set_matplotlib_formats
            <https://github.com/ipython/matplotlib-inline/blob/b93777db35267acefe6e37d14214360362d2e8b2/matplotlib_inline/backend_inline.py#L280-L281>`_
            for details.

        """
        if cls._is_run_from_ipython():
            # No docs yet: https://github.com/ipython/matplotlib-inline/issues/12
            from matplotlib_inline.backend_inline import set_matplotlib_formats

            if isinstance(ipython_format, str):
                ipython_format = [ipython_format]

            set_matplotlib_formats(*ipython_format)

        from matplotlib import rcParams

        cls._vector_friendly = vector_friendly
        cls.file_format_figs = format
        if dpi is not None:
            rcParams["figure.dpi"] = dpi
        if dpi_save is not None:
            rcParams["savefig.dpi"] = dpi_save
        if transparent is not None:
            rcParams["savefig.transparent"] = transparent
        if facecolor is not None:
            rcParams["figure.facecolor"] = facecolor
            rcParams["axes.facecolor"] = facecolor
        if scanpy:
            from ..plotting._rcmod import set_rcParams_scanpy

            set_rcParams_scanpy(fontsize=fontsize, color_map=color_map)
        if figsize is not None:
            rcParams["figure.figsize"] = figsize
        cls._frameon = frameon

    @staticmethod
    def _is_run_from_ipython():
        """Determine whether we're currently in IPython."""
        import builtins

        return getattr(builtins, "__IPYTHON__", False)

    @classmethod
    def _default_logfile(cls) -> TextIO:
        return sys.stdout if cls._is_run_from_ipython() else sys.stderr

    def __str__(cls) -> str:
        return "\n".join(
            f"{k} = {v!r}"
            for k, v in inspect.getmembers(cls)
            if not k.startswith("_") and k != "getdoc"
        )


class settings(metaclass=SettingsMeta):  # noqa: N801
    """Settings for scanpy."""

    def __new__(cls) -> type[Self]:
        return cls

    # logging
    _root_logger: ClassVar = _RootLogger(logging.WARNING)
    _logfile: ClassVar = SettingsMeta._default_logfile()
    _logpath: ClassVar = None
    _verbosity: ClassVar = Verbosity.warning
    # rest
    _n_pcs: ClassVar = 50
    _plot_suffix: ClassVar = ""
    _file_format_data: ClassVar = "h5ad"
    _file_format_figs: ClassVar = "pdf"
    _autosave: ClassVar = False
    _autoshow: ClassVar = True
    _writedir: ClassVar = Path("./write")
    _cachedir: ClassVar = Path("./cache")
    _datasetdir: ClassVar = Path("./data")
    _figdir: ClassVar = Path("./figures")
    _cache_compression: ClassVar = "lzf"
    _max_memory: ClassVar = 15
    _n_jobs: ClassVar = 1
    _categories_to_ignore: ClassVar = ["N/A", "dontknow", "no_gate", "?"]
    _frameon: ClassVar = True
    _vector_friendly: ClassVar = False
    _low_resolution_warning: ClassVar = True
    _start: ClassVar = time()
    _previous_time: ClassVar = _start
    _previous_memory_usage: ClassVar = -1


if not documenting():  # finish initialization
    _set_log_level(settings, settings.verbosity.level)
    _set_log_file(settings)



from __future__ import annotations

from collections.abc import MutableMapping
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from packaging.version import Version
from sklearn.utils import check_random_state

from .. import logging as logg
from .._compat import CSBase, old_positionals, pkg_version
from .._settings import settings
from .._utils import NeighborsView, raise_not_implemented_error_if_backed_type
from .._utils._doctests import doctest_skip
from ..neighbors import FlatTree

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

    from anndata import AnnData

    from ..neighbors import RPForestDict

ANNDATA_MIN_VERSION = Version("0.7rc1")


@old_positionals(
    "obs",
    "embedding_method",
    "labeling_method",
    "neighbors_key",
    "neighbors_key",
    "inplace",
)
@doctest_skip("illustrative short example but not runnable")
def ingest(
    adata: AnnData,
    adata_ref: AnnData,
    *,
    obs: str | Iterable[str] | None = None,
    embedding_method: str | Iterable[str] = ("umap", "pca"),
    labeling_method: str = "knn",
    neighbors_key: str | None = None,
    inplace: bool = True,
    **kwargs,
):
    """Map labels and embeddings from reference data to new data.

    :doc:`/tutorials/basics/integrating-data-using-ingest`

    Integrates embeddings and annotations of an `adata` with a reference dataset
    `adata_ref` through projecting on a PCA (or alternate
    model) that has been fitted on the reference data. The function uses a knn
    classifier for mapping labels and the UMAP package :cite:p:`McInnes2018` for mapping
    the embeddings.

    .. note::

        We refer to this *asymmetric* dataset integration as *ingesting*
        annotations from reference data to new data. This is different from
        learning a joint representation that integrates both datasets in an
        unbiased way, as CCA (e.g. in Seurat) or a conditional VAE (e.g. in
        scVI) would do.

    You need to run :func:`~scanpy.pp.neighbors` on `adata_ref` before
    passing it.

    Parameters
    ----------
    adata
        The annotated data matrix of shape `n_obs` × `n_vars`. Rows correspond
        to cells and columns to genes. This is the dataset without labels and
        embeddings.
    adata_ref
        The annotated data matrix of shape `n_obs` × `n_vars`. Rows correspond
        to cells and columns to genes.
        Variables (`n_vars` and `var_names`) of `adata_ref` should be the same
        as in `adata`.
        This is the dataset with labels and embeddings
        which need to be mapped to `adata`.
    obs
        Labels' keys in `adata_ref.obs` which need to be mapped to `adata.obs`
        (inferred for observation of `adata`).
    embedding_method
        Embeddings in `adata_ref` which need to be mapped to `adata`.
        The only supported values are 'umap' and 'pca'.
    labeling_method
        The method to map labels in `adata_ref.obs` to `adata.obs`.
        The only supported value is 'knn'.
    neighbors_key
        If not specified, ingest looks at adata_ref.uns['neighbors']
        for neighbors settings and adata_ref.obsp['distances'] for
        distances (default storage places for pp.neighbors).
        If specified, ingest looks at adata_ref.uns[neighbors_key] for
        neighbors settings and
        adata_ref.obsp[adata_ref.uns[neighbors_key]['distances_key']] for distances.
    inplace
        Only works if `return_joint=False`.
        Add labels and embeddings to the passed `adata` (if `True`)
        or return a copy of `adata` with mapped embeddings and labels.

    Returns
    -------
    Returns `None` if `copy=False`, else returns an `AnnData` object. Sets the following fields:

    `adata.obs[obs]` : :class:`pandas.Series` (dtype ``category``)
        Mapped labels.
    `adata.obsm['X_umap' | 'X_pca']` : :class:`numpy.ndarray` (dtype ``float``)
        Mapped embeddings. `'X_umap'` if `embedding_method` is `'umap'`, `'X_pca'` if `embedding_method` is `'pca'`.

    Example
    -------
    Call sequence:

    >>> import scanpy as sc
    >>> sc.pp.neighbors(adata_ref)
    >>> sc.tl.umap(adata_ref)
    >>> sc.tl.ingest(adata, adata_ref, obs="cell_type")

    """
    # anndata version check
    anndata_version = pkg_version("anndata")
    if anndata_version < ANNDATA_MIN_VERSION:
        msg = (
            f"ingest only works correctly with anndata>={ANNDATA_MIN_VERSION} "
            f"(you have {anndata_version}) as prior to {ANNDATA_MIN_VERSION}, "
            "`AnnData.concatenate` did not concatenate `.obsm`."
        )
        raise ValueError(msg)

    start = logg.info("running ingest")
    obs = [obs] if isinstance(obs, str) else obs
    embedding_method = (
        [embedding_method] if isinstance(embedding_method, str) else embedding_method
    )
    labeling_method = (
        [labeling_method] if isinstance(labeling_method, str) else labeling_method
    )

    if len(labeling_method) == 1 and len(obs or []) > 1:
        labeling_method = labeling_method * len(obs)

    ing = Ingest(adata_ref, neighbors_key)
    ing.fit(adata)

    for method in embedding_method:
        ing.map_embedding(method)

    if obs is not None:
        ing.neighbors(**kwargs)
        for i, col in enumerate(obs):
            ing.map_labels(col, labeling_method[i])

    logg.info("    finished", time=start)
    return ing.to_adata(inplace=inplace)


def _rp_forest_generate(
    rp_forest_dict: RPForestDict,
) -> Generator[FlatTree, None, None]:
    props = FlatTree._fields
    num_trees = len(rp_forest_dict[props[0]]["start"]) - 1

    for i in range(num_trees):
        tree = []
        for prop in props:
            start = rp_forest_dict[prop]["start"][i]
            end = rp_forest_dict[prop]["start"][i + 1]
            tree.append(rp_forest_dict[prop]["data"][start:end])
        yield FlatTree(*tree)

    tree = []
    for prop in props:
        start = rp_forest_dict[prop]["start"][num_trees]
        tree.append(rp_forest_dict[prop]["data"][start:])
    yield FlatTree(*tree)


class _DimDict(MutableMapping):
    def __init__(self, dim, axis=0, vals=None):
        self._data = {}
        self._dim = dim
        self._axis = axis
        if vals is not None:
            self.update(vals)

    def __setitem__(self, key, value):
        if value.shape[self._axis] != self._dim:
            msg = (
                f"Value passed for key {key!r} is of incorrect shape. "
                f"Value has shape {value.shape[self._axis]} "
                f"for dimension {self._axis} while "
                f"it should have {self._dim}."
            )
            raise ValueError(msg)
        self._data[key] = value

    def __getitem__(self, key):
        return self._data[key]

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return f"{type(self).__name__}({self._data})"


class Ingest:
    """Class to map labels and embeddings from existing data to new data.

    You need to run :func:`~scanpy.pp.neighbors` on `adata` before
    initializing Ingest with it.

    Parameters
    ----------
    adata : :class:`~anndata.AnnData`
        The annotated data matrix of shape `n_obs` × `n_vars`
        with embeddings and labels.

    """

    def _init_umap(self, adata):
        from umap import UMAP

        self._umap = UMAP(
            metric=self._metric,
            random_state=adata.uns["umap"]["params"].get("random_state", 0),
        )

        self._umap._initial_alpha = self._umap.learning_rate
        self._umap._raw_data = self._rep
        self._umap.knn_dists = None

        self._umap._validate_parameters()

        self._umap.embedding_ = adata.obsm["X_umap"]
        self._umap._sparse_data = isinstance(self._rep, CSBase)
        self._umap._small_data = self._rep.shape[0] < 4096
        self._umap._metric_kwds = self._metric_kwds

        self._umap._n_neighbors = self._n_neighbors
        self._umap.n_neighbors = self._n_neighbors

        self._umap._knn_search_index = self._nnd_idx

        self._umap._a = adata.uns["umap"]["params"]["a"]
        self._umap._b = adata.uns["umap"]["params"]["b"]

        self._umap._input_hash = None

    def _init_pynndescent(self, distances):
        from pynndescent import NNDescent

        first_col = np.arange(distances.shape[0])[:, None]
        init_indices = np.hstack((first_col, np.stack(distances.tolil().rows)))

        self._nnd_idx = NNDescent(
            data=self._rep,
            metric=self._metric,
            metric_kwds=self._metric_kwds,
            n_neighbors=self._n_neighbors,
            init_graph=init_indices,
            random_state=self._neigh_random_state,
        )

        # temporary hack for the broken forest storage
        from pynndescent.rp_trees import make_forest

        current_random_state = check_random_state(self._nnd_idx.random_state)
        self._nnd_idx._rp_forest = make_forest(
            self._nnd_idx._raw_data,
            self._nnd_idx.n_neighbors,
            self._nnd_idx.n_search_trees,
            self._nnd_idx.leaf_size,
            self._nnd_idx.rng_state,
            current_random_state,
            self._nnd_idx.n_jobs,
            self._nnd_idx._angular_trees,
        )

    def _init_neighbors(self, adata, neighbors_key):
        neighbors = NeighborsView(adata, neighbors_key)

        self._n_neighbors = neighbors["params"]["n_neighbors"]

        if "use_rep" in neighbors["params"]:
            self._use_rep = neighbors["params"]["use_rep"]
            self._rep = adata.X if self._use_rep == "X" else adata.obsm[self._use_rep]
        elif "n_pcs" in neighbors["params"]:
            self._use_rep = "X_pca"
            self._n_pcs = neighbors["params"]["n_pcs"]
            self._rep = adata.obsm["X_pca"][:, : self._n_pcs]
        elif adata.n_vars > settings.N_PCS and "X_pca" in adata.obsm:
            self._use_rep = "X_pca"
            self._rep = adata.obsm["X_pca"][:, : settings.N_PCS]
            self._n_pcs = self._rep.shape[1]

        self._metric_kwds = neighbors["params"].get("metric_kwds", {})
        self._metric = neighbors["params"]["metric"]

        self._neigh_random_state = neighbors["params"].get("random_state", 0)
        self._init_pynndescent(neighbors["distances"])

    def _init_pca(self, adata):
        self._pca_centered = adata.uns["pca"]["params"]["zero_center"]
        self._pca_use_hvg = adata.uns["pca"]["params"]["use_highly_variable"]

        mask = "highly_variable"
        if self._pca_use_hvg and mask not in adata.var.columns:
            msg = f"Did not find `adata.var[{mask!r}']`."
            raise ValueError(msg)

        if self._pca_use_hvg:
            self._pca_basis = adata.varm["PCs"][adata.var[mask]]
        else:
            self._pca_basis = adata.varm["PCs"]

    def __init__(self, adata: AnnData, neighbors_key: str | None = None):
        # assume rep is X if all initializations fail to identify it
        self._rep = adata.X
        self._use_rep = "X"

        self._n_pcs = None

        self._adata_ref = adata
        self._adata_new: AnnData | None = None

        if "pca" in adata.uns:
            self._init_pca(adata)

        if neighbors_key is None:
            neighbors_key = "neighbors"

        if neighbors_key in adata.uns:
            self._init_neighbors(adata, neighbors_key)
        else:
            msg = (
                f'There is no neighbors data in `adata.uns["{neighbors_key}"]`.\n'
                "Please run pp.neighbors."
            )
            raise ValueError(msg)

        if "X_umap" in adata.obsm:
            self._init_umap(adata)

        self._obsm = None
        self._obs = None
        self._labels = None

        self._indices = None
        self._distances = None

    def _pca(self, n_pcs=None):
        x = self._adata_new.X
        x = x.toarray() if isinstance(x, CSBase) else x.copy()
        if self._pca_use_hvg:
            x = x[:, self._adata_ref.var["highly_variable"]]
        if self._pca_centered:
            x -= x.mean(axis=0)
        x_pca = np.dot(x, self._pca_basis[:, :n_pcs])
        return x_pca

    def _same_rep(self):
        adata = self._adata_new
        if self._n_pcs is not None:
            return self._pca(self._n_pcs)
        if self._use_rep == "X":
            return adata.X
        if self._use_rep in adata.obsm:
            return adata.obsm[self._use_rep]
        return adata.X

    def fit(self, adata_new):
        """Map `adata_new` to the same representation as `adata`.

        This function identifies the representation which was used to
        calculate neighbors in 'adata' and maps `adata_new` to
        this representation.
        Variables (`n_vars` and `var_names`) of `adata_new` should be the same
        as in `adata`.

        `adata` refers to the :class:`~anndata.AnnData` object
        that is passed during the initialization of an Ingest instance.
        """
        raise_not_implemented_error_if_backed_type(adata_new.X, "Ingest.fit")
        ref_var_names = self._adata_ref.var_names.str.upper()
        new_var_names = adata_new.var_names.str.upper()

        if not ref_var_names.equals(new_var_names):
            msg = (
                "Variables in the new adata are different "
                "from variables in the reference adata"
            )
            raise ValueError(msg)

        self._obs = pd.DataFrame(index=adata_new.obs.index)
        self._obsm = _DimDict(adata_new.n_obs, axis=0)

        self._adata_new = adata_new
        self._obsm["rep"] = self._same_rep()

    def neighbors(self, k=None, queue_size=5, epsilon=0.1, random_state=0):
        """Calculate neighbors of `adata_new` observations in `adata`.

        This function calculates `k` neighbors in `adata` for
        each observation of `adata_new`.
        """
        from umap.umap_ import INT32_MAX, INT32_MIN

        random_state = check_random_state(random_state)
        rng_state = random_state.randint(INT32_MIN, INT32_MAX, 3).astype(np.int64)

        test = self._obsm["rep"]

        if k is None:
            k = self._n_neighbors

        self._nnd_idx.search_rng_state = rng_state
        self._indices, self._distances = self._nnd_idx.query(test, k, epsilon)

    def _umap_transform(self):
        return self._umap.transform(self._obsm["rep"])

    def map_embedding(self, method):
        """Map embeddings of `adata` to `adata_new`.

        This function infers embeddings, specified by `method`,
        for `adata_new` from existing embeddings in `adata`.
        `method` can be 'umap' or 'pca'.
        """
        if method == "umap":
            self._obsm["X_umap"] = self._umap_transform()
        elif method == "pca":
            self._obsm["X_pca"] = self._pca()
        else:
            msg = "Ingest supports only umap and pca embeddings for now."
            raise NotImplementedError(msg)

    def _knn_classify(self, labels):
        # ensure it's categorical
        cat_array: pd.Series = self._adata_ref.obs[labels].astype("category")
        values = [cat_array.iloc[inds].mode()[0] for inds in self._indices]
        return pd.Categorical(values=values, categories=cat_array.cat.categories)

    def map_labels(self, labels, method):
        """Map labels of `adata` to `adata_new`.

        This function infers `labels` for `adata_new.obs`
        from existing labels in `adata.obs`.
        `method` can be only 'knn'.
        """
        if method == "knn":
            self._obs[labels] = self._knn_classify(labels)
        else:
            msg = "Ingest supports knn labeling for now."
            raise NotImplementedError(msg)

    @old_positionals("inplace")
    def to_adata(self, *, inplace: bool = False) -> AnnData | None:
        """Return `adata_new` with mapped embeddings and labels.

        If `inplace=False` returns a copy of `adata_new`
        with mapped embeddings and labels in `obsm` and `obs` correspondingly.
        If `inplace=True` returns nothing and updates `adata_new.obsm`
        and `adata_new.obs` with mapped embeddings and labels.
        """
        adata = self._adata_new if inplace else self._adata_new.copy()

        adata.obsm.update(self._obsm)

        for key in self._obs:
            adata.obs[key] = self._obs[key]

        if not inplace:
            return adata

    def to_adata_joint(
        self, batch_key="batch", batch_categories=None, index_unique="-"
    ):
        """Return concatenated object.

        This function returns the new :class:`~anndata.AnnData` object
        with concatenated existing embeddings and labels of 'adata'
        and inferred embeddings and labels for `adata_new`.
        """
        adata = self._adata_ref.concatenate(
            self._adata_new,
            batch_key=batch_key,
            batch_categories=batch_categories,
            index_unique=index_unique,
        )

        obs_update = self._obs.copy()
        obs_update.index = adata[adata.obs[batch_key] == "1"].obs_names
        adata.obs.update(obs_update)

        for key in self._obsm:
            if key in self._adata_ref.obsm:
                adata.obsm[key] = np.vstack((
                    self._adata_ref.obsm[key],
                    self._obsm[key],
                ))

        if self._use_rep not in ("X_pca", "X"):
            adata.obsm[self._use_rep] = np.vstack((
                self._adata_ref.obsm[self._use_rep],
                self._obsm["rep"],
            ))

        if "X_umap" in self._obsm:
            adata.uns["umap"] = self._adata_ref.uns["umap"]
        if "X_pca" in self._obsm:
            adata.uns["pca"] = self._adata_ref.uns["pca"]
            adata.varm["PCs"] = self._adata_ref.varm["PCs"]

        return adata
